# coding: utf-8
"""
Classes and functions for parsing the ONCVPSP output file and plotting the results.
"""
from __future__ import annotations

import io
import os
import json
import tempfile
import numpy as np

from typing import Any, Union, Optional
from shutil import which
from monty.collections import dict2namedtuple
from monty.termcolor import cprint
from scipy.interpolate import UnivariateSpline
from abipy.core.atom import l2char # NlkState, RadialFunction, RadialWaveFunction,
from abipy.core.mixins import NotebookWriter
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_visible, set_axlims
from abipy.tools.typing import Figure
from abipy.tools.derivatives import finite_diff
from abipy.ppcodes.oncv_parser import OncvParser


class OncvPlotter(NotebookWriter):
    """
    Plots the results produced by a pseudopotential generator.
    """
    # TODO: Improve support for fully-relativistic case.

    # matplotlib options.
    linewidth, markersize = 2, 2

    linestyle_aeps = dict(ae="solid", ps="dashed")

    markers_aeps = dict(ae=".", ps="o")

    color_l = {0: "black",
               1: "red",
               -1: "magenta",
               2: "blue",
               -2: "cyan",
               3: "orange",
               -3: "yellow",
    }

    @classmethod
    def from_file(cls, filepath: str) -> OncvPlotter:
        """Build the plotter from an output file."""
        parser = OncvParser(filepath)
        parser.scan()
        if not parser.run_completed:
            raise RuntimeError("oncvpsp output is not completed")

        return cls(parser)

    def __init__(self, parser: OncvParser):
        self.parser = parser

    @staticmethod
    def decorate_ax(ax, xlabel=None, ylabel=None, title=None, fontsize=8):
        """
        Decorate a `matplotlib` Axis adding xlabel, ylabel, title, grid and legend
        """
        if title: ax.set_title(title, fontsize=fontsize)
        if xlabel: ax.set_xlabel(xlabel)
        if ylabel: ax.set_ylabel(ylabel)
        ax.grid(True)
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

    def _mpl_opts_laeps(self, l: int, aeps: str) -> dict:
        """
        Return dict with matplotlib ax.plot options to plot AE/PS quantities depending on l.
        """
        return dict(
            color=self.color_l[l],
            linestyle=self.linestyle_aeps[aeps],
            linewidth=self.linewidth,
            #marker=self.markers_aeps[aeps],
            markersize=self.markersize
        )

    def _add_rc_vlines_ax(self, ax, with_lloc=False) -> None:
        """
        Add vertical lines to axis `ax` showing the core radii.
        """
        for l, rc in self.parser.rc_l.items():
            ax.axvline(rc, lw=2, color=self.color_l[l], ls="--")

        if with_lloc:
            color = "magenta" if self.parser.lloc == 4 else "k"
            ax.axvline(self.parser.rc5, lw=2, color=color, ls="--")

    @add_fig_kwargs
    def plot_atan_logders(self, ax=None, with_xlabel=True,
                          fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot arctan of logder on axis ax.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ae, ps = self.parser.atan_logders.ae, self.parser.atan_logders.ps
        ax, fig, plt = get_ax_fig_plt(ax)

        # Note that l can be negative if FR pseudo. This corresponds to ikap 2 in Fortran.

        for l, ae_alog in ae.items():
            ps_alog = ps[l]

            if not self.parser.relativistic:
                lch = f"${l2char[abs(l)]}$"
            else:
                #lch = l2char[abs(l)]
                lch = f"${l2char[abs(l)]}^+$" if l >= 0 else f"${l2char[abs(l)]}^-$"

            # Add pad to avoid overlapping curves. We only need to compare AE vs PS atan(logder)
            pad = (abs(l) + 1) * 1.0

            ae_line, = ax.plot(ae_alog.energies, ae_alog.values + pad,
                               label=f"AE {lch}",
                               **self._mpl_opts_laeps(l, "ae"))

            ps_line, = ax.plot(ps_alog.energies, ps_alog.values + pad,
                               label=f"PS {lch}",
                               **self._mpl_opts_laeps(l, "ps"))

        xlabel = "Energy (Ha)" if with_xlabel else ""
        #ylabel = "ATAN(LogDer)"
        ylabel = r"$\phi(E) = \arctan(R * d \psi_E(r)/dr |_R)$"

        self.decorate_ax(ax, xlabel=xlabel, ylabel=ylabel, title="",
                         fontsize=fontsize,
                         )
        return fig

    def _get_ae_ps_wfs(self, what) -> tuple:
        if what == "bound_states":
            ae_wfs, ps_wfs = self.parser.radial_wfs.ae, self.parser.radial_wfs.ps
        elif what == "scattering_states":
            ae_wfs, ps_wfs = self.parser.scattering_wfs.ae, self.parser.scattering_wfs.ps
        else:
            raise ValueError(f"Invalid value for {what=}")
        return ae_wfs, ps_wfs


    @add_fig_kwargs
    def plot_radial_wfs(self, ax=None, what="bound_states",
                        fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot AE and PS radial wavefunctions on axis ax.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            what: "bound_states" or "scattering_states".
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        ae_wfs, ps_wfs = self._get_ae_ps_wfs(what)

        for nlk, ae_wf in ae_wfs.items():
            ps_wf, l, k = ps_wfs[nlk], nlk.l, nlk.k

            if what == "bound_states":
                # Show position of the last peak.
                s, marker = 10, "^"
                ae_peaks = ae_wf.get_peaks()
                style = dict(color=self.color_l[l], s=s, marker=marker)
                ax.scatter(ae_peaks.xs[-1], ae_peaks.ys[-1], **style)
                ps_peaks = ps_wf.get_peaks()
                style = dict(color=self.color_l[l], s=s, marker=marker)
                ax.scatter(ps_peaks.xs[-1], ps_peaks.ys[-1], **style)

            ax.plot(ae_wf.rmesh, ae_wf.values, label=f"AE {nlk.latex}", **self._mpl_opts_laeps(l, "ae"))
            ax.plot(ps_wf.rmesh, ps_wf.values, label=f"PS {nlk.latex}", **self._mpl_opts_laeps(l, "ps"))

        self.decorate_ax(ax, xlabel="r (Bohr)", ylabel=r"$\phi(r)$",
                         title="Wave Functions" if what == "bound_states" else "Scattering States",
                         fontsize=fontsize,
                         )

        self._add_rc_vlines_ax(ax)

        return fig

    @add_fig_kwargs
    def plot_projectors(self, ax=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot projectors on axis ax.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        linestyle_n = {1: "solid", 2: "dashed", 3: "dotted", 4: "dashdot"}

        for nlk, proj in self.parser.projectors.items():
            ax.plot(proj.rmesh, proj.values,
                    color=self.color_l.get(nlk.l, 'black'),
                    linestyle=linestyle_n[nlk.n],
                    linewidth=self.linewidth,
                    markersize=self.markersize,
                    label=f"Proj {nlk.n}, l={nlk.latex_l}",
                    )

        self.decorate_ax(ax, xlabel="r (Bohr)", ylabel="$p(r)$", title="Projectors",
                         fontsize=fontsize,
                         )

        self._add_rc_vlines_ax(ax)

        return fig

    @add_fig_kwargs
    def plot_densities(self, ax=None, timesr2=False, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot AE, PS and model densities on axis ax.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        for name, rho in self.parser.densities.items():
            d = rho.values if not timesr2 else rho.values * rho.rmesh ** 2
            line, = ax.plot(rho.rmesh, d, label=name,
                            linewidth=self.linewidth, markersize=self.markersize)

        ylabel = "$n(r)$" if not timesr2 else "$r^2 n(r)$"
        self.decorate_ax(ax, xlabel="r (Bohr)", ylabel=ylabel,
                         title="Charge densities",
                         fontsize=fontsize,
                         )
        return fig

    @add_fig_kwargs
    def plot_der_densities(self, ax=None, order=1, acc=4, fontsize=8, **kwargs) -> Figure:
        """
        Plot the radial derivatives of the densities on axis ax.
        Used to analyze possible discontinuities or strong oscillations in r-space.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        for name, rho in self.parser.densities.items():
            # Only model core charge is shown.
            if name != "rhoM": continue

            # Need linear mesh for finite_difference --> Spline input densities on lin_rmesh
            lin_rmesh, h = np.linspace(rho.rmesh[0], rho.rmesh[-1], num=len(rho.rmesh) * 4, retstep=True)
            spline = UnivariateSpline(rho.rmesh, rho.values, s=0)
            lin_values = spline(lin_rmesh)
            vder = finite_diff(lin_values, h, order=order, acc=acc)
            ax.plot(lin_rmesh, vder, label="%s-order derivative of %s" % (order, name))

        self.decorate_ax(ax, xlabel="r (Bohr)", ylabel="$D^%s \n(r)$" % order,
                         title="Derivative of the charge densities",
                         fontsize=fontsize,
                         )
        return fig

    @add_fig_kwargs
    def plot_potentials(self, ax=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot v_l and v_loc potentials on axis ax.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        for l, pot in self.parser.potentials.items():
            ax.plot(pot.rmesh, pot.values,
                    label="$V_{loc}$" if l == -1 else "PS $V_{%s}$" % l2char[l],
                    **self._mpl_opts_laeps(l, "ae"))

        self.decorate_ax(ax, xlabel="r (Bohr)", ylabel="$v_l(r)$",
                         title="Ion Pseudopotentials", fontsize=fontsize,
                         )
        self._add_rc_vlines_ax(ax, with_lloc=True)

        return fig

    @add_fig_kwargs
    def plot_der_potentials(self, ax=None, order=1, acc=4, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot the derivatives of vl and vloc potentials on axis ax.
        Used to analyze the derivative discontinuity introduced by the RRKJ method at rc.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        for l, pot in self.parser.potentials.items():
            # Need linear mesh for finite_difference hence spline input potentials on lin_rmesh.
            lin_rmesh, h = np.linspace(pot.rmesh[0], pot.rmesh[-1], num=len(pot.rmesh) * 4, retstep=True)
            spline = UnivariateSpline(pot.rmesh, pot.values, s=None)
            lin_values = spline(lin_rmesh)
            vder = finite_diff(lin_values, h, order=order, acc=acc)

            label = f"{order}-order derivative of Vloc" if l == -1 else \
                    f"{order}-order derivative of PS l={l}"

            ax.plot(lin_rmesh, vder, label=label, **self._mpl_opts_laeps(l, "ae"))

        self.decorate_ax(ax, xlabel="r (Bohr)", ylabel=r"$D^%s \phi(r)$" % order,
                         title="Derivative of the ion Pseudopotentials",
                         fontsize=fontsize,
                         )
        self._add_rc_vlines_ax(ax, with_lloc=True)

        return fig

    @add_fig_kwargs
    def plot_kene_vs_ecut(self, ax=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot the convergence of the kinetic energy wrt ecut on axis ax.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        for l, data in self.parser.kene_vs_ecut.items():
            ax.plot(data.energies, data.values, label="Conv l=%s" % l2char[l],
                    **self._mpl_opts_laeps(l, "ae"))

        for nlk, data in self.parser.kinerr_nlk.items():
            ax.plot(data.ecuts, data.values_ha,
                    **self._mpl_opts_laeps(nlk.l, "ps"))

        self.decorate_ax(ax,
                         xlabel="Ecut (Ha)", ylabel=r"$\Delta E_{kin}$ (Ha)",
                         title="", fontsize=fontsize,
                         )

        ax.set_yscale("log")

        return fig

    @add_fig_kwargs
    def plot_atanlogder_econv(self, ax_list=None, fontsize: int = 6, **kwargs) -> Figure:
        """
        Plot atan(logder) and the convergence of kinetic energy on the same figure.

        Return: matplotlib Figure
        """
        # Build grid of plots.
        ax_list, fig, plt = get_axarray_fig_plt(ax_list, nrows=2, ncols=1,
                                                sharex=False, sharey=False, squeeze=False)
        ax_list = ax_list.ravel()

        self.plot_atan_logders(ax=ax_list[0], fontsize=fontsize, show=False)
        ax_list[0].xaxis.set_label_position('top')

        self.plot_kene_vs_ecut(ax=ax_list[1], fontsize=fontsize, show=False)

        return fig

    @add_fig_kwargs
    def plot_den_formfact(self, ecut=120, ax=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot the density form factor as a function of ecut in Ha.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.

        Return: matplotlib Figure.
        """
        ax, fig, plt = get_ax_fig_plt(ax)

        for name, rho in self.parser.densities.items():
            if name == "rhoC": continue
            form = rho.get_intr2j0(ecut=ecut) / (4 * np.pi)
            ax.plot(form.mesh, form.values, label=name,
                    linewidth=self.linewidth, markersize=self.markersize)

            intg = rho.r2f_integral()[-1]
            #print("r2 f integral: ", intg)
            #print("form_factor(0): ", name, form.values[0])

        # Plot vloc(q)
        #for l, pot in self.potentials.items():
        #    if l != -1: continue
        #    form = pot.get_intr2j0(ecut=ecut)
        #    mask = np.where(np.abs(form.values) > 20); form.values[mask] = 20
        #    line, = ax.plot(form.mesh, form.values, linewidth=self.linewidth, markersize=self.markersize)
        #    lines.append(line); legends.append("Vloc(q)")

        self.decorate_ax(ax,
                         xlabel="Ecut (Ha)", ylabel="$n(q)$",
                         fontsize=fontsize, title="Form factor, l=0 ",
                         )
        return fig

    @add_fig_kwargs
    def plot_atomic_levels(self, ax=None, fontsize: int = 0, **kwargs) -> Figure:
        """
        Plot atomic levels in Ha.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.

        Return: matplotlib Figure.
        """
        max_eval = max(e.eig for e in self.parser.atomic_levels if e.is_valence)
        min_eval = min(e.eig for e in self.parser.atomic_levels if e.is_valence)
        min_ene = min_eval - 1.0

        # Filter the atomic_levels to be printed
        atm_levels = [l for l in self.parser.atomic_levels if l.eig >= min_ene]
        xs = [1] * len(atm_levels)
        enes = [level.eig for level in atm_levels]

        figsize = (6.5, 12)
        figsize = None
        ax, fig, plt = get_ax_fig_plt(ax, figsize=figsize)
        ax.scatter(xs, enes, s=90000, marker="_", linewidth=2, zorder=3)

        for xi, yi, level in zip(xs, enes, atm_levels):
            text = f"{level.nlk.latex}" + f", occ={level.occ}"
            #xy = .65*xi, yi
            #xy = xi, yi
            xy = xi, 0.95 * yi
            ax.annotate(text, xy=xy, xytext=(8, 4), size=8, ha="center", va='top', textcoords="offset points")

        span_style = {}
        span_style.setdefault("alpha", 0.2)
        span_style.setdefault("color", "grey")
        rectangle = ax.axhspan(min_eval, max_eval, **span_style)

        #import matplotlib.patches as patches
        #p1 = patches.FancyArrowPatch((0, 0), (1, 1), arrowstyle='<->', mutation_scale=20)
        #p2 = patches.FancyArrowPatch((1, 0), (0, 1), arrowstyle='<|-|>', mutation_scale=20)

        #ax.margins(0.1)
        ax.set_ylabel('Eigenvalue (Ha)')
        ax.set_title('Atomic energy levels')
        ax.set_xticks([])
        #ax.yaxis.set_minor_locator(mpl.ticker.MaxNLocator(50))
        ax.grid(axis='y')
        #ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        Generate a predefined list of matplotlib figures with minimal input from the user.
        """
        verbose = kwargs.get("verbose", 0)

        yield self.plot_atanlogder_econv(show=False)
        yield self.plot_potentials(show=False)
        yield self.plot_der_potentials(show=False)
        yield self.plot_radial_wfs(show=False)
        if self.parser.has_scattering_wfs:
            yield self.plot_radial_wfs(what="scattering_states", show=False)
        yield self.plot_projectors(show=False)
        yield self.plot_densities(show=False)
        #yield self.plot_densities(timesr2=True, show=False)
        yield self.plot_den_formfact(show=False)
        yield self.plot_atomic_levels(show=False)
        if verbose:
            #yield self.plot_der_potentials(show=False)
            for order in [1, 2, 3, 4]:
                yield self.plot_der_densities(order=order, show=False)

    def write_notebook(self, nbpath=None):
        return oncv_make_open_notebook(self.parser.filepath)


def oncv_make_open_notebook(outpath: str,
                            foreground: bool = False,
                            classic_notebook: bool = False,
                            no_browser: bool = False) -> int:  # pragma: no cover
    """
    Generate an ipython notebook and open it in the browser.

    Args:
        foreground: By default, jupyter is executed in background and stdout, stderr are redirected
            to devnull. Use foreground to run the process in foreground
        classic_notebook: True to use the classic notebook instead of jupyter-lab (default)
        no_browser: Start the jupyter server to serve the notebook but don't open the notebook in the browser.
            Use this option to connect remotely from localhost to the machine running the kernel

    Return: system exit code.

    Raise: `RuntimeError` if jupyter executable is not in $PATH
    """
    nbpath = oncv_write_notebook(outpath, nbpath=None)

    if not classic_notebook:
        # Use jupyter-lab.
        app_path = which("jupyter-lab")
        if app_path is None:
            raise RuntimeError("""
Cannot find jupyter-lab application in $PATH. Install it with:

    conda install -c conda-forge jupyterlab

or:

    pip install jupyterlab

See also https://jupyterlab.readthedocs.io/
""")

    else:
        # Use classic notebook
        app_path = which("jupyter")
        if app_path is None:
            raise RuntimeError("""
Cannot find jupyter application in $PATH. Install it with:

    conda install -c conda-forge jupyter

or:

    pip install jupyterlab

See also https://jupyter.readthedocs.io/en/latest/install.html
""")
        app_path = app_path + " notebook "

    if not no_browser:
        if foreground:
            return os.system("%s %s" % (app_path, nbpath))
        else:
            fd, tmpname = tempfile.mkstemp(text=True)
            print(tmpname)
            cmd = "%s %s" % (app_path, nbpath)
            print("Executing:", cmd, "\nstdout and stderr redirected to %s" % tmpname)
            import subprocess
            process = subprocess.Popen(cmd.split(), shell=False, stdout=fd, stderr=fd)
            cprint("pid: %s" % str(process.pid), "yellow")
            return 0

    else:
        # Based on https://github.com/arose/nglview/blob/master/nglview/scripts/nglview.py
        notebook_name = os.path.basename(nbpath)
        dirname = os.path.dirname(nbpath)
        #print("nbpath:", nbpath)

        import socket
        from abipy.tools.notebooks import find_free_port
        username = os.getlogin()
        hostname = socket.gethostname()
        port = find_free_port()

        client_cmd = "ssh -NL localhost:{port}:localhost:{port} {username}@{hostname}".format(
            username=username, hostname=hostname, port=port)

        print(f"""
Using port: {port}

\033[32m In your local machine, run: \033[0m

                {client_cmd}

\033[32m NOTE: you might want to replace {hostname} by full hostname with domain name \033[0m
\033[32m Then open your web browser, copy and paste the URL: \033[0m

http://localhost:{port}/notebooks/{notebook_name}
""")
        if not classic_notebook:
            cmd = f'{app_path} {notebook_name} --no-browser --port {port} --notebook-dir {dirname}'
        else:
            cmd = f'{app_path} notebook {notebook_name} --no-browser --port {port} --notebook-dir {dirname}'

        print("Executing:", cmd)
        print('NOTE: make sure to open `{}` in your local machine\n'.format(notebook_name))

        return os.system(cmd)


def oncv_write_notebook(outpath: str, nbpath: Optional[str] = None) -> str:
    """
    Write an ipython notebook to nbpath
    If nbpath is None, a temporay file is created.
    Return path to the notebook.
    """
    outpath = os.path.abspath(outpath)

    import nbformat
    nbf = nbformat.v4
    nb = nbf.new_notebook()

    nb.cells.extend([
        nbf.new_markdown_cell("## This is an auto-generated notebook for %s" % os.path.basename(outpath)),
        nbf.new_code_cell("""\
%matplotlib notebook

# Use this magic for jupyterlab.
# For installation instructions, see https://github.com/matplotlib/jupyter-matplotlib
#%matplotlib widget

"""),

        nbf.new_code_cell("""\
# Parse output file
from abipy.ppcodes.oncv_parser import OncvParser
onc_parser = OncvParser('%s')""" % outpath),

        nbf.new_code_cell("""\
# Parse the file and build the plotter
onc_parser.scan()
if not onc_parser.run_completed:
    raise RuntimeError("Cannot parse output file")

plotter = onc_parser.get_plotter()"""),

        nbf.new_markdown_cell(r"# AE and PS radial wavefunctions $\phi(r)$:"),
        nbf.new_code_cell("fig = plotter.plot_radial_wfs(show=False)"),
        nbf.new_markdown_cell("# Arctan of the logarithmic derivatives:"),
        nbf.new_code_cell("fig = plotter.plot_atan_logders(show=False)"),
        nbf.new_markdown_cell("# Convergence in $G$-space estimated by ONCVPSP:"),
        nbf.new_code_cell("fig = plotter.plot_kene_vs_ecut(show=False)"),
        nbf.new_markdown_cell("# Projectors:"),
        nbf.new_code_cell("fig = plotter.plot_projectors(show=False)"),
        nbf.new_markdown_cell("# Core-Valence-Model charge densities:"),
        nbf.new_code_cell("fig = plotter.plot_densities(show=False)"),
        nbf.new_markdown_cell("# Local potential and $l$-dependent potentials:"),
        nbf.new_code_cell("fig = plotter.plot_potentials(show=False)"),

        #nbf.new_markdown_cell("# 1-st order derivative of $v_l$ and $v_{loc}$ computed via finite differences:"),
        #nbf.new_code_cell("""fig = plotter.plot_der_potentials(order=1, show=False)"""),
        #nbf.new_markdown_cell("# 2-nd order derivative of $v_l$ and $v_{loc}$ computed via finite differences:"),
        #nbf.new_code_cell("""fig = plotter.plot_der_potentials(order=2, show=False)"""),
        #nbf.new_markdown_cell("Model core charge and form factors computed by ABINIT"),
        #nbf.new_code_cell("""\
#with pseudo.open_pspsfile() as psps:
#psps.plot()"""),
    ])

    # Plot data
    #plotter.plot_der_potentials()
    #for order in [1,2,3,4]:
    #    plotter.plot_der_densities(order=order)
    #plotter.plot_densities(timesr2=True)
    #plotter.plot_den_formfact()

    if nbpath is None:
        _, nbpath = tempfile.mkstemp(suffix='.ipynb', text=True)

    with io.open(nbpath, 'wt', encoding="utf8") as f:
        nbformat.write(nb, f)

    return nbpath


class MultiOncvPlotter(NotebookWriter):
    """
    Object for comparing and plotting multiple pseudos.

    Usage example:

    .. code-block:: python

        plotter = MultiOncvPlotter.from_files(filepaths)
        plotter.plot_atan_logders()
    """

    @classmethod
    def from_files(cls, files: list[str]) -> MultiOncvPlotter:
        """
        Create an instance from a list of oncvpsp output files.
        """
        new = cls()
        for file in files:
            new.add_file(file, file)

        return new

    def __init__(self):
        self._plotters_dict = {}

    def add_file(self, label: str, filepath: str) -> None:
        """
        Add a oncvps output file to the plotter with label
        """
        if label in self._plotters_dict:
            raise ValueError(f"Cannot overwrite label: {label}")

        parser = OncvParser(filepath)
        parser.scan()
        plotter = parser.get_plotter()
        if plotter is not None:
            self._plotters_dict[label] = plotter

    def __len__(self) -> int:
        return len(self._plotters_dict)

    @property
    def plotters(self) -> list[OncvPlotter]:
        """"List of registered `Plotters`."""
        return list(self._plotters_dict.values())

    @property
    def labels(self) -> list[str]:
        """List of labels."""
        return list(self._plotters_dict.keys())

    def items(self):
        """List of plotters."""
        return self._plotters_dict.items()

    def _get_ax_list(self, ax_list, sharex=True, sharey=False, ravel=True, layout="c"):

        if layout == "c":
            num_plots, ncols, nrows = len(self), 1, len(self)
        elif layout == "r":
            num_plots, ncols, nrows = len(self), len(self), 1
        else:
            raise ValueError(f"Invalid {layout=}")

        # Build grid of plots.
        ax_list, fig, plt = get_axarray_fig_plt(ax_list, nrows=nrows, ncols=ncols, # figsize=(8, 8),
                                                sharex=sharex, sharey=sharey, squeeze=False)
        if ravel:
            ax_list = ax_list.ravel()

        return ax_list, fig, plt

    @add_fig_kwargs
    def plot_atan_logders(self, ax_list=None, with_xlabel=True, xlims=None, ylims=None,
                          fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot arctan of logder on ax_list for all pseudos.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=True)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_atan_logders(ax=ax, with_xlabel=with_xlabel, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            set_axlims(ax, xlims, "x")
            set_axlims(ax, ylims, "y")
            if i != len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_radial_wfs(self, ax_list=None, what="bound_states",
                        fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot AE and PS radial wavefunctions of ax_list for all pseudos.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
            what: "bound_states" or "scattering_states".
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=True)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_radial_wfs(ax=ax, what=what, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            if i != len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_projectors(self, ax_list=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot projectors on ax_list for all pseudos.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=True)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_projectors(ax=ax, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            if i != len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_densities(self, ax_list=None, timesr2=False, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot AE, PS and model densities on ax_list for all pseudos.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
            ax: |matplotlib-Axes| or None if a new figure should be created.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=False)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_densities(ax=ax, timesr2=timesr2, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            if i != len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_der_densities(self, ax_list=None, order=1, acc=4, fontsize=8, **kwargs) -> Figure:
        """
        Plot the radial derivatives of the densities on ax_list for all pseudos.
        Used to analyze possible discontinuities or strong oscillations in r-space.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=False)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_der_densities(ax=ax, order=order, acc=acc, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            if i != len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_potentials(self, ax_list=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot v_l and v_loc potentials on ax_list for all pseudos.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=False)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_potentials(ax=ax, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            if i !=  len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_der_potentials(self, ax_list=None, order=1, acc=4, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot the derivatives of vl and vloc potentials on ax_list for all pseudos.
        Used to analyze the derivative discontinuity introduced by the RRKJ method at rc.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=False)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_der_potentials(ax=ax, order=order, acc=4, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            if i != len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_kene_vs_ecut(self, ax_list=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot the convergence of the kinetic energy wrt ecut on ax_list for all pseudos.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=True)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_kene_vs_ecut(ax=ax, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            if i != len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_atanlogder_econv(self, ax_mat=None, fontsize: int = 6, **kwargs) -> Figure:
        """
        Plot atan(logder) and converge of kinetic energy on the same figure for all pseudos.
        Return: matplotlib Figure
        """
        num_plots, ncols, nrows = 2 * len(self), 2, len(self)

        # Build grid of plots.
        ax_mat, fig, plt = get_axarray_fig_plt(ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)

        for i, (label, plotter) in enumerate(self.items()):
            ax_list = ax_mat[i]
            n = len(ax_list)
            plotter.plot_atanlogder_econv(ax_list=ax_list, fontsize=fontsize, show=False)
            for ax in ax_list:
                ax.set_title(label, fontsize=fontsize)
            if i != n - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_den_formfact(self, ecut=120, ax_list=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot the density form factor as a function of ecut in Ha on ax_list for all pseudos.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.

        Return: matplotlib Figure.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=False)

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_den_formfact(ax=ax, ecut=ecut, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            if i != len(ax_list) - 1:
                set_visible(ax, False, "legend", "xlabel", "ylabel")

        return fig

    @add_fig_kwargs
    def plot_atomic_levels(self, ax_list=None, fontsize: int = 8, **kwargs) -> Figure:
        """
        Plot atomic levels in Ha for all pseudos.

        Args:
            ax_list: List of |matplotlib-Axes| or None if a new figure should be created.

        Return: matplotlib Figure.
        """
        ax_list, fig, plt = self._get_ax_list(ax_list, sharex=False, sharey=True, layout="r")

        for i, (ax, (label, plotter)) in enumerate(zip(ax_list, self.items())):
            plotter.plot_atomic_levels(ax=ax, fontsize=fontsize, show=False)
            ax.set_title(label, fontsize=fontsize)
            #if i != len(ax_list) - 1:
            #    set_visible(ax, False, "legend", "xlabel", "ylabel")

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        Generate a predefined list of matplotlib figures with minimal input from the user.
        """
        verbose = kwargs.get("verbose", 0)

        yield self.plot_atan_logders(show=False)
        yield self.plot_kene_vs_ecut(show=False)
        yield self.plot_radial_wfs(show=False)
        if any(plotter.parser.has_scattering_wfs for plotter in self.plotters):
            yield self.plot_radial_wfs(what="scattering_states", show=False)
        yield self.plot_potentials(show=False)
        #if verbose:
        yield self.plot_der_potentials(show=False)
        yield self.plot_projectors(show=False)
        yield self.plot_densities(show=False)
        ##yield self.plot_densities(timesr2=True, show=False)
        yield self.plot_den_formfact(show=False)
        yield self.plot_atomic_levels(show=False)
        #if verbose:
        #    for order in [1, 2, 3, 4]:
        #        yield self.plot_der_densities(order=order, show=False)

    def write_notebook(self, nbpath=None):
        raise NotImplementedError("write_notebooks should be tested")
        #return oncv_make_open_notebook(self.parser.filepath)


def psp8_get_densities(path, fc_file=None, ae_file=None, plot=False):
    """
    Extract AE-core, AE-valence and PS-valence from a psp8 file.
    Optionally write `.fc` and `.AE` file in format suitable for AIM and cut3d Hirshfeld code.

    Args:
        path: path of the psp8 file
        fc_file, ae_file: File-like object to `.fc.` and `.AE` file
            Set to None if files are now wanted.
        plot: If true, call matplotlib to plot densities.

    Return:
        namedtuple with numpy arrays: (rmesh, psval, aeval, aecore)
            Densities are multiplied by 4 pi i.e. int aecore * r**2 dr ~= N_core

    .. warning::

        The densities do not integrate exactly to the number of core/valence electrons.
        in particular the valence charge in elements with large Z as valence electrons are
        more delocalized and the finite radial mesh does not capture the contribution due to the tail.
        Client code is responsible for extrapolating the valence charge if needed.
        The error in the integral of the core charge is usually much smaller and should be ok
        for qualitative analysis.
    """
    # From 64_psp/psp8in.F90
    # (1) title (character) line
    # (2) znucl,zion,pspdat
    # (3) pspcod,pspxc,lmax,lloc,mmax,r2well  (r2well not used)
    # (4) rchrg,fchrg,qchrg  (fchrg /=0 if core charge, qchrg not used)
    # (5) nproj(0:lmax)  (several projectors allowed for each l)
    # Then, for ll=0,lmax:
    # if nproj(ll)>0
    #   1/<u1|vbkb1>, 1/<u2|vbkb2>, ...
    #   for irad=1,mmax: irad, r(irad), vbkb1(irad,ll), vbkb2(irad,ll), ...
    # elif ll=lloc
    #   for  irad=1,mmax: irad, r(irad), vloc(irad)
    # end if
    #
    # If (lloc>lmax)
    # for  irad=1,mmax: irad, r(irad), vloc(irad)
    # end if
    #
    # vbkb are Bloechl-Kleinman-Bylander projectors,(vpsp(r,ll)-vloc(r))*u(r,ll),
    # unnormalized
    # Note that an arbitrary local potential is allowed.  Set lloc>lmax, and
    # provide projectors for all ll<=lmax
    #
    # Finally, if fchrg>0,
    # for  irad=1,mmax: irad, r(irad), xccc(irad),
    # xccc'(irac), xccc''(irad), xccc'''(irad), xccc''''(irad)
    #
    # Model core charge for nonlinear core xc correction, and 4 derivatives
    from abipy.flowtk.pseudos import Pseudo
    pseudo = Pseudo.from_file(path)

    from pymatgen.io.abinit.pseudos import _dict_from_lines
    with open(path, "rt") as fh:
        lines = [fh.readline() for _ in range(6)]

        header = _dict_from_lines(lines[1:3], [3, 6])

        # Number of points on the linear mesh.
        mmax = int(header["mmax"])

        # Non-linear core correction parameters.
        rchrg, fchrg, qchrg = [float(t) for t in lines[3].split()[:3]]

        # Read Number of projectors(l) and extension switch
        nproj = [int(t) for t in lines[4].split()[:5]]
        assert len(nproj) == 5
        tokens = lines[5].split()
        extension_switch = int(tokens[0])

        # Old format, Densities are not available.
        if len(tokens) == 1 or tokens[1] == "extension_switch":
            raise RuntimeError("psp8 file does not contain density records")

        den_flag = int(tokens[1])
        if den_flag != 1:
            raise ValueError("Expecting den_flag 1 but got %s" % den_flag)

        # Read SOC projectors
        has_soc = extension_switch in (2, 3)
        if has_soc:
            line = fh.readline()
            # Start at l=1
            nproj_soc = [int(t) for t in line.split()[:4]]
            nproj_soc.insert(0, 0)
            #print("nproj_soc", nproj_soc)
            raise NotImplementedError("SOC not tested")

        lmax = int(header["lmax"])
        lloc = int(header["lloc"])
        nso = 1 if not has_soc else 2

        # Will now proceed at the reading of pots and projectors
        # rad(:)=radial grid r(i)
        # vpspll(:,1),...,vpspll(:,lnmax)=nonlocal projectors
        # vloc(:)=local potential

        #for nn in range(nso):
        # Skip projectors (scalar relativistic, always present).
        for l, npl in enumerate(nproj):
            #if npl == 0 and l != lloc: continue
            if npl == 0: continue
            line = fh.readline() # l, ekb[:npl]
            l_file = int(line.split()[0])
            if l != l_file:
                #print("For l=%s, npl=%s" % (l, npl), "wrong line", line)
                raise RuntimeError("l != l_file (%s != %s)" % (l, l_file))

            for ir in range(mmax):
                line = fh.readline()
                assert int(line.split()[0]) == ir + 1

        # Skip local potential.
        if lloc == 4:
            lloc_file = int(fh.readline())
            assert lloc_file == lloc
            for ir in range(mmax):
                fh.readline()

        # Skip model core charge function and derivatives, if present.
        if fchrg > 1e-15:
            for ir in range(mmax):
                fh.readline()

        # Read pseudo valence charge in real space on the linear mesh.
        # [i, r, PS_val, AE_val, AE_core]
        rmesh, psval, aeval, aecore = [np.empty(mmax) for _ in range(4)]
        for ir in range(mmax):
            l = fh.readline()
            #print("denline", l)
            findx, rad, v1, v2, v3 = l.split()
            assert ir + 1 == int(findx)
            rmesh[ir] = float(rad.replace("D", "E"))
            psval[ir] = float(v1.replace("D", "E"))
            aeval[ir] = float(v2.replace("D", "E"))
            aecore[ir] = float(v3.replace("D", "E"))

        #fact = 1 / (4 * np.pi)
        #aeval *= fact
        #psval *= fact
        #aecore *= fact
        from scipy.integrate import simps
        r2 = rmesh ** 2

        meta = dict(
            aeval_integral=simps(aeval * r2, x=rmesh),
            psval_integral=simps(psval * r2, x=rmesh),
            aecore_integral=simps(aecore * r2, x=rmesh),
            aeden_integral=simps((aecore + aeval) * r2, x=rmesh),
            symbol=pseudo.symbol,
            Z=pseudo.Z,
            Z_val=pseudo.Z_val,
            l_max=pseudo.l_max,
            md5=pseudo.md5,
        )

        if fc_file is not None:
            # Write results to file in "fc" format. The files contains:
            # header with number of points and unknown parameter (not used)
            # for each point in the radial mesh:
            #    4 columns with the radial r coordinate, the core density at r,
            #    and the first and second derivatives of the core density.
            # See http://www.abinit.org/downloads/core_electron
            #aeval_spline = UnivariateSpline(rmesh, aeval)
            #psval_spline = UnivariateSpline(rmesh, psval)
            aecore_spline = UnivariateSpline(rmesh, aecore)
            f1 = aecore_spline.derivative(1)
            f2 = aecore_spline.derivative(2)
            header = "%d 0.0 %s %s %s # nr, dummy, symbol, Z, Z_val" % (
                mmax, pseudo.symbol, pseudo.Z, pseudo.Z_val)
            print(header, file=fc_file)
            for ir in range(mmax):
                r = rmesh[ir]
                print(4 * "%.14E  " % (r, aecore[ir], f1(r), f2(r)), file=fc_file)

            print("\n\n<JSON>", file=fc_file)
            print(json.dumps(meta, indent=4), file=fc_file)
            print("</JSON>", file=fc_file)

        if ae_file is not None:
            # Write results to file in "AE" format
            # The files contains:
            #   header with number of points and unknown parameter (not used)
            #   then 2 columns with the radial r coordinate and the AE density at r
            #   See http://www.abinit.org/downloads/all_core_electron
            header = "%d 0.0 %s %s %s # nr, dummy, symbol, Z, Z_val" % (
                mmax, pseudo.symbol, pseudo.Z, pseudo.Z_val)
            print(header, file=ae_file)
            for ir in range(mmax):
                print(2 * "%.14E  " % (rmesh[ir], aecore[ir] + aeval[ir]), file=ae_file)

            print("\n\n<JSON>", file=ae_file)
            print(json.dumps(meta, indent=4), file=ae_file)
            print("</JSON>", file=ae_file)

        if plot:
            import matplotlib.pyplot as plt
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1)
            ax.plot(rmesh, r2 * aecore, label="AE core * r**2")
            ax.plot(rmesh, r2 * psval, label="PS valence * r**2")
            ax.plot(rmesh, r2 * aeval, label="AE valence * r**2")
            ax.grid(True)
            ax.legend(loc="best")
            plt.show()

        return dict2namedtuple(rmesh=rmesh, psval=psval, aeval=aeval, aecore=aecore)



#!/usr/bin/env python
r"""
Thermodinamic properties
========================

This example shows how to compute and plot thermodinamic properties within
the harmonic approximation using the phonon DOS produced by anaddb.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

#%%
# Read the Phonon DOS from the netcd file produced by anaddb (prtdos 2)

ncfile = abiopen(abidata.ref_file("trf2_5.out_PHDOS.nc"))
phdos = ncfile.phdos

#%%
# Print crystalline structure and zero-point energy.

print(ncfile.structure)
zpe = phdos.zero_point_energy
print("Zero point energy:", zpe, zpe.to("J"), zpe.to("Ha"))

#%%
# Compute free energy from 2 to 300 K (20 points)
# By default, energies are is eV and thermodynamic quantities are given
# on a per-unit-cell basis.

f = phdos.get_free_energy(tstart=2, tstop=300, num=20)
#f.plot()

#%%
# Plot U, F, S, Cv as a function of T.
# Use J/mol units, results are divided by formula_units.

phdos.plot_harmonic_thermo(units="Jmol", formula_units=1)

#%%
# Plotly version with dark template.

phdos.plotly_harmonic_thermo(units="Jmol", formula_units=1,
                             title=f"Thermodinamic properties of {ncfile.structure.formula}",
                             template="plotly_dark",
                             )

#%%
# Remember to close the file:
ncfile.close()



from __future__ import annotations

#import panel as pn

from typing import List
from pydantic import Field, root_validator
#from abipy.abio.inputs import AbinitInput
#from abipy.panels.core import ply
#from abipy.panels.viewers import JSONViewer
#from abipy.flowtk import TaskManager, Flow
from .base_models import MongoConnector  # , GfsFileDesc
from .flow_models import FlowModel, PresetQuery



class GbrvFlowModel(FlowModel):

    #scf_input: AbinitInput = Field(None, description="Abinit Input file generated by AbiPy.")

    ########
    # Output
    ########

    #scf_data: ScfData = Field(None, description="Results produced by the SCF run.")

    def build_flow(self, workdir: str, worker: AbipyWorker) -> Flow:
        """
        Build an AbiPy Flow in workdir using the input data available in the model and return it.
        """
        pseudos = self.pseudos_specs.get_pseudos()
        #return flow

    def postprocess_flow(self, flow: Flow, mng_connector: MongoConnector) -> None:
        """
        Analyze the flow and fills the model with output results.
        MongoConnector should be used only to insert files in GridFs as the final insertion is done by the caller.
        This function is called by the AbiPy Worker if the flow completed successfully.
        """

    def get_panel_view(self, mng_connector: MongoConnector):
        """
        Return panel object with a view of the model.
        """

    @classmethod
    def get_preset_queries(cls) -> List[PresetQuery]:
        """
        Return list of dictionaries with the MongoDB queries typically used to filter documents for this model.
        Empty list if no suggestion is available.
        """
        #return [
        #    PresetQuery(
        #        query={"$or": [
        #            {"scf_data.abs_pressure_gpa": {"$gt": p_gpa}},
        #            {"scf_data.max_force_ev_over_ang": {"$gt": 1e-4}}]
        #        },
        #        projection=["in_structure_data.formula_pretty",
        #                    "scf_data.abs_pressure_gpa", "scf_data.max_force_ev_over_ang",
        #                   ],
        #        info=f"Filter `{cls.__name__}` documents with pressure > {p_gpa} GPa "
        #             f"or absolute max force > {max_force} eV/Ang",
        #        )
        #]




"""
This module defines convenience types for type hinting purposes.
It extends the types provided by pymatgen with Abipy-specific ones.
"""
from __future__ import annotations

import numpy as np

from typing import TYPE_CHECKING, Any, Union, Sequence


if TYPE_CHECKING:  # needed to avoid circular imports
    from matplotlib.pyplot import Axes
    from matplotlib.figure import Figure
    from abipy.core.kpoints import Kpoint
else:
    Axes = Any
    Figure = Any
    Kpoint = Any

VectorLike = Union[Sequence[float], np.ndarray]

# matplotlib objects
#AxList = list[Axes]

# Abipy objects
KptLike = Union["Kpoint", VectorLike]
KptSelect = Union[int, "Kpoint", "VectorLike"]




"""
This Module defines the AbipyWorker object
"""
from __future__ import annotations

import sys
import os
import threading
import json
import socket
import time
import subprocess
import tempfile
import getpass
import functools
import tornado
import tornado.web
import tornado.escape
import pandas as pd
import requests
import panel as pn

from datetime import datetime
from uuid import uuid4
#from queue import Queue, Empty
from typing import Union, Optional  # , Type  # Awaitable
from enum import Enum
from monty.json import MSONable
from monty.termcolor import colored
#from monty.functools import lazy_property
from tqdm import tqdm

from abipy.tools.serialization import pmg_serialize
from abipy.tools.printing import print_dataframe
from abipy.tools import duck
from abipy.tools.iotools import yaml_safe_load_path
from abipy.flowtk.flows import Flow
from abipy.flowtk.tasks import TaskManager, set_user_config_taskmanager
from abipy.flowtk.launcher import MultiFlowScheduler
from .tools import pid_exists, port_is_open, find_free_port
from .base_models import AbipyModel, Field, MongoConnector, AbipyEncoder


# This should become a global variable used in all the other modules to facilitate unit tests.
ABIPY_DIRPATH = os.path.join(os.path.expanduser("~"), ".abinit", "abipy")


class BaseHandler(tornado.web.RequestHandler):
    """
    Our base class for RequestHandler used by the AbipyWorker bokeh server.
    Each handler has a reference to the worker set in `initialize`.
    """

    #def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
    #    pass

    def initialize(self, worker: AbipyWorker):
        self.worker = worker

    def get_json_from_body(self):
        try:
            data = tornado.escape.json_decode(self.request.body)
            return data
        except ValueError:
            # TODO: handle the error
            raise

    #def write_error(self, status_code, **kwargs):
    #    # https://stackoverflow.com/questions/26371051/better-way-to-handle-errors-in-tornado-request-handler
    #   import traceback, logging
    #   if status_code == 500:
    #      excp = kwargs['exc_info'][1]
    #      tb   = kwargs['exc_info'][2]
    #      stack = traceback.extract_tb(tb)
    #      clean_stack = [i for i in stack if i[0][-6:] != 'gen.py' and i[0][-13:] != 'concurrent.py']
    #      error_msg = '{}\n  Exception: {}'.format(''.join(traceback.format_list(clean_stack)),excp)

    #      # do something with this error now... e.g., send it to yourself
    #      # on slack, or log it.
    #      logging.error(error_msg)  # do something with your error...

    #   # don't forget to show a user friendly error page!
    #   self.render("oops.html")


class ActionHandler(BaseHandler):

    SUPPORTED_METHODS = ("POST",)

    def post(self):
        data = self.get_json_from_body()
        action = data.get("action", None)
        if action is None: return
        reply = {}

        if action == "kill":
            reply["message"] = "Received kill action. Aborting now!"
            self.write(reply)
            # TODO: Should lock and give a change to the secondary thread to complete operations
            time.sleep(4)
            sys.exit(0)

        #elif action == "rm_errored_flows":
        #    self.worker.scheduler.rm_flows_with_status("Error")
        #    reply["message"] = f"Removed: {nflows} Errored Flows"

        else:
            reply["message"] = f"Unknown action: {action}"

        self.write(reply)


class PostFlowHandler(BaseHandler):
    """
    This endpoint allows the client to send a python script to the server.
    """

    SUPPORTED_METHODS = ("POST",)

    def post(self):
        #if self.worker.flow_model is not None: raise RuntimeError("Cannot send stuff in flow_model is ON")
        data = self.get_json_from_body()

        pyscript_basename = data["pyscript_basename"]
        pyscript_string = data["pyscript_string"]
        user_message = data["user_message"]

        scratch_dir = self.worker.scratch_dir
        filepath = os.path.join(scratch_dir, pyscript_basename)
        with open(filepath, "wt") as fh:
            fh.write(pyscript_string)

        now = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        workdir = tempfile.mkdtemp(prefix=f"Flow-{now}", dir=scratch_dir)

        # FIXME: problem with manager and possible collisions in filepath
        p = subprocess.Popen(f"python {filepath} -w {workdir}",
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        out, err = p.communicate()

        reply = dict(
            returncode=p.returncode,
            out=out.decode("utf-8") if out is not None else None,
            err=err.decode("utf-8") if err is not None else None,
            workdir=workdir,
        )

        if p.returncode == 0:
            print(f"Running new Flow generated by {pyscript_basename} in workdir: {workdir}")
            flow = Flow.from_file(workdir)
            self.worker.flow_scheduler.add_flow(flow, user_message)

        self.write(reply)


class _JsonHandler(BaseHandler):

    def prepare(self):
        self.set_header(name="Content-Type", value="application/json")


class JsonStatusHandler(_JsonHandler):
    """
    This endpoint allows the client to get a json document with the status of the server.
    """

    SUPPORTED_METHODS = ("GET",)

    def get(self):
        # FIXME: This is broken now
        json_status = self.worker.flow_scheduler.get_json_status()
        print("json_status", json_status)
        self.write(json_status)


class FlowDirsPostHandler(BaseHandler):
    """
    This endpoint allows the client to send to the server a list of directories
    containing pregenerated Flows. It is assumed that both the client and the server
    are running on the same host machine.
    """

    SUPPORTED_METHODS = ("POST",)

    def post(self):
        data = self.get_json_from_body()
        user_message = data["user_message"]
        flow_dir = data["flow_dir"]

        error = ""
        try:
            flow = Flow.from_file(flow_dir)
            self.worker.flow_scheduler.add_flow(flow, user_message)
        except Exception as exc:
            error = str(exc)

        reply = dict(
            returncode=0 if not error else 1,
            error=error,
        )

        self.write(reply)


class WorkerStatusEnum(str, Enum):
    """
    Possible status of an AbipyWorker instance.
    """
    init = "init"
    serving = "serving"
    dead = "dead"

    def colored_str(self) -> str:
        s = self.value
        color = {"init": "blue", "serving": "green",  "dead": "red"}[s]
        return colored(s, color=color)


def _str_uuid() -> str:
    return str(uuid4())


class WorkerState(AbipyModel):

    name: str = Field(..., description="Name of the AbiPy worker. It is assumed to be unique.")

    status: WorkerStatusEnum = Field(WorkerStatusEnum.init, description="Status of the worker.")

    pid: int = Field(default_factory=os.getpid, description="Process identifier")

    address: str = Field("localhost", description="Server address.")

    port: int = Field(0, description="Server port. 0 means find a free port at runtime.")

    scratch_dir: str = Field(..., description="Scratch directory used by the Worker to generate Flows.")

    username: str = Field(default_factory=getpass.getuser, description="User name at the host")

    hostname: str = Field(default_factory=socket.gethostname, description="Host name.")

    #uuid: UUID = Field(default_factory=uuid4)
    uuid: str = Field(default_factory=_str_uuid,
                      description="AbipyWorker ID. Used by the client to detect if the server has restarted.")

    mng_connector: MongoConnector = Field(None, description="MongoDB connection parameters and name of default collection")



#class FlowContext(BaseModel):
#    workdir: str
#    manager: TaskManager
#    mng_connector: MongoConnector


class AbipyWorker:
    """
    An AbipyWorker is a bokeh web-server running on the same machine where Abipy Flows are executed.
    An AbipyWorker has a name that is supposed to be unique across different machines.
    Flows are scheduled according to the `sched_options` and executed inside the `scratch_dir` directory.

    A command line interface to create/start/restart Workers is provided by the `abiw.py` script.
    """
    MAX_NUM_EXCS_IN_BUILDFLOW = 30

    #MONGO_PERIODIC_CALLBACK_TIME_MS = 10000
    MONGO_PERIODIC_CALLBACK_TIME_MS = 5000

    def __init__(self, name: str, sched_options: dict, scratch_dir: str,
                 address: str = "localhost", port: int = 0,
                 manager: Optional[TaskManager] = None,
                 mng_connector: Optional[MongoConnector] = None):
        """
        Args:
            name: The name of the AbipyWorker. Must be unique across machines.
            sched_options:
            scratch_dir:
            address:
                The address the server should listen on for HTTP requests.
            port:
                 Allows specifying a specific port
                 Port 0 means to select an arbitrary unused port
            manager:
            mng_connector:
        """
        self.name = name
        self.address, self.port = address, port
        self.pid = os.getpid()
        self.manager = manager
        self.mng_connector = mng_connector
        self.flow_model_cls = None
        self.num_exceptions_in_build_flow_from_mongodb = 0
        self.flowid2_oid_model = {}

        if self.mng_connector:
            # Get the FlowMode subclass from the DB collection.
            collection = self.mng_connector.get_collection()
            from .flow_models import FlowModel
            self.flow_model_cls = FlowModel.get_subclass_from_collection(collection)

        # url --> callables returning panel objects.
        self.routes = {
            "/": self.serve_worker_homepage,
            r"/flow/\d+": self.serve_panel_flow,
        }

        # url --> tornado handler
        self.extra_patterns = [
            ("/json_status", JsonStatusHandler, dict(worker=self)),
            ("/action", ActionHandler, dict(worker=self)),
        ]

        if self.flow_model_cls is None:
            # These endpoints are activated only if we are not using a MongoDB collection.
            self.extra_patterns.extend([
                ("/post_flow_script", PostFlowHandler, dict(worker=self)),
                ("/post_flow_dirs", FlowDirsPostHandler, dict(worker=self)),
            ])

        self.config_dir = os.path.join(ABIPY_DIRPATH, f"worker_{self.name}")
        if not os.path.exists(self.config_dir):
            os.mkdir(self.config_dir)

        self.scratch_dir = scratch_dir
        if not os.path.isdir(scratch_dir):
            raise ValueError(f"Scratch directory: `{scratch_dir}` does not exist!")

        sqldb_path = os.path.join(self.config_dir, "flows.db")
        self.flow_scheduler = MultiFlowScheduler(sqldb_path, **sched_options)

        state_file = os.path.join(self.config_dir, "state.json")
        if os.path.exists(state_file):
            state = WorkerState.from_json_file(state_file)
            if state.status == state.status.serving:
                pid = state.pid
                if pid_exists(pid):
                    raise RuntimeError(f"There's already an AbiPyWorker serving on this machine with pid: {pid}")
                # FIXME This does not work in demon mode!
                print(f"Status 'serving' but pid {pid} does not exist!")

        #self.write_state_file(status="init")

    # TODO
    #@lazy_property
    #def full_name(self)
    #    return f"{self.name}@{socket.gethostname()}"

    def __repr__(self) -> str:
        return f"{self.name} at {self.address}:{self.port}. pid: {self.pid}"

    #def __str__(self) -> str:
    #    lines = []
    #    app = lines.append
    #    return "\n".join(lines)

    #def mkworkdir(self):
    #    now = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
    #    workdir = tempfile.mkdtemp(prefix=f"Flow-{now}", dir=self.scratch_dir)

    def build_flow_from_mongodb(self) -> None:
        # Stop the server if the number of consecutive exceptions becomes too large
        # This may be due to e.g. a quota limit or other hardware-related problems.
        #print("in build_flow")
        try:
            self._build_flow_from_mongodb()
            if self.num_exceptions_in_build_flow_from_mongodb > 0:
                self.num_exceptions_in_build_flow_from_mongodb -= 1

        except Exception as exc:
            print(exc)
            self.num_exceptions_in_build_flow_from_mongodb += 1

        if self.num_exceptions_in_build_flow_from_mongodb > self.MAX_NUM_EXCS_IN_BUILDFLOW:
            self.shutdown_server("Too many exceptions in build_flow_from_mongodb")

    def _build_flow_from_mongodb(self) -> None:
        """
        Periodic callback executed by the primary thread to fetch new calculations from
        the MongoDB collection when then AbipyWorker is running with a `flow_model_cls`
        """
        if self.flow_model_cls is None:
            raise ValueError("The AbipyWorker should have a flow_model_cls when running with a MongDB database.")

        # TODO: Improve error handling.
        collection = self.mng_connector.get_collection()

        if self.flowid2_oid_model:
            node_ids = list(self.flowid2_oid_model.keys())
            rows = self.flow_scheduler.get_sql_rows_with_node_ids(node_ids)
            #print("Found rows", rows)

            for row in rows:
                flow_id = row["flow_id"]
                if row["status"] == "Completed":
                    print("In completed")
                    flow = Flow.from_file(row["workdir"])
                    oid, model = self.flowid2_oid_model[flow_id]
                    model.postprocess_flow_and_update_db(flow, oid, self)
                    self.flowid2_oid_model.pop(flow_id)
                else:
                    pass
                    #print("Should handle row with status", row["status"])

        # TODO: Implement Pipelines?
        # Use atomic update of the models so that we can have multiple workers possibly on
        # different machines fetching and updating documents in the same collection.
        #print("Finding runnable_oid_models")
        oid_flowmodel_list = self.flow_model_cls.find_runnable_oid_models(collection, limit=5)
        if not oid_flowmodel_list: return
        print("in build_flow_from_mongodb, found", len(oid_flowmodel_list), "flows in database")

        for oid, model in oid_flowmodel_list:
            #print("Running model:", repr(model))
            now = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
            workdir = tempfile.mkdtemp(prefix=f"Flow-{now}", dir=self.scratch_dir)

            flow = model.build_flow_and_update_collection(workdir, oid, self)
            if flow is None: continue
            self.flow_scheduler.add_flow(flow, user_message="")  # m.user_message ?
            self.flowid2_oid_model[flow.node_id] = (oid, model)

    @classmethod
    def _get_state_path(cls, name: str) -> tuple[WorkerState, str]:
        config_dir = os.path.join(ABIPY_DIRPATH, f"worker_{name}")
        state_filepath = os.path.join(config_dir, "state.json")

        return WorkerState.from_json_file(state_filepath), state_filepath

    @classmethod
    def init_from_dirname(cls, dirname: str) -> AbipyWorker:
        """
        Initialize the AbipyWorker from a directory basename.
        """
        state, path = cls._get_state_path(dirname)

        if state.status != state.status.init:
            raise RuntimeError(f"`status` entry in json file: `{path}`\n"
                               f"should be set to `init` while it is: `{state.status}`")

        config_dir = os.path.dirname(path)
        sched_options = yaml_safe_load_path(os.path.join(config_dir, "scheduler.yml"))
        manager = TaskManager.from_file(os.path.join(config_dir, "manager.yml"))

        # TODO: mng_connector, flow_model_cls
        return cls(state.name, sched_options, state.scratch_dir,
                   address=state.address, port=state.port, manager=manager,
                   mng_connector=state.mng_connector)

    @classmethod
    def new_with_name(cls, worker_name: str, scratch_dir: str,
                      scheduler_path: Optional[str] = None,
                      manager_path: Optional[str] = None,
                      mng_connector=None, verbose: int = 1) -> AbipyWorker:
        """
        Create a new AbiPyWorker.

        Args:
            worker_name:
            scratch_dir:
            scheduler_path:
            manager_path:
            mng_connector:
            verbose:
        """
        config_dir = os.path.join(ABIPY_DIRPATH, f"worker_{worker_name}")
        errors = []
        eapp = errors.append

        if os.path.exists(config_dir):
            eapp(f"Directory `{config_dir}` already exists!")

        scheduler_path = scheduler_path or os.path.join(ABIPY_DIRPATH, "scheduler.yml")
        manager_path = manager_path or os.path.join(ABIPY_DIRPATH, "manager.yml")

        if not os.path.exists(scheduler_path):
            eapp(f"Cannot find scheduler file at `{scheduler_path}` to initialize the worker!")
        if not os.path.exists(manager_path):
            eapp(f"Cannot find manager file at `{manager_path}` to initialize the worker!")
        if not os.path.isdir(scratch_dir):
            eapp(f"scratch_dir: `{scratch_dir}` is not a directory!")

        if errors:
            lines = ["The following problems have been detected:"]
            for i, err in enumerate(errors):
                lines.append(f"\t [{i+1}] {err}")
            raise RuntimeError("\n".join(lines))

        if verbose:
            print(f"""
Creating new worker directory `{config_dir}`
Copying manager.yml and scheduler.yml files into it.
These are the configuration files used by the AbipyWorker to submit and schedule tasks.
You may want to customize these files before running calculations.

After this step, use:

    abiw.py start {worker_name}

to start the AbiPy worker, and:

    abiw.py lscan

to update the list of local clients.
""")

        os.mkdir(config_dir)
        from shutil import copy
        copy(scheduler_path, config_dir)
        copy(manager_path, config_dir)

        worker_state = WorkerState(
                name=worker_name,
                status=WorkerStatusEnum.init,
                scratch_dir=scratch_dir,
                mng_connector=mng_connector,
        )

        worker_state.json_write(os.path.join(config_dir, "state.json"), indent=2)

        return cls.init_from_dirname(worker_name)

    @classmethod
    def restart_from_dirname(cls, name: str, force: bool = False) -> AbipyWorker:
        """
        Restart the AbipyWorker given its name. Must be executed on the same machine
        where the Worker resides.
        """
        try:
            state, path = cls._get_state_path(name)
        except FileNotFoundError as exc:
            print(exc)
            raise ValueError(f"{name} is a remote Worker.\n"
                             "To restart it, execute the command on the same machine where the worker resides.")

        pid = state.pid

        #if state.status != "dead":
        # TODO: status can be set to serving if the worker has been killed by kill -9
        # should implement a restart_hard option and perhaps make a bkp of the database!
        if state.status == state.status.serving:
            e_pid = pid_exists(pid)

            msg = f"It seems there's already a worker serving on this machine with pid: {pid}"
            if not force:
                if not e_pid:
                    print(f"Status is set to `serving` but pid {pid} does not exist!")
                else:
                    raise RuntimeError(msg)
            else:
                print(msg)
                print("but as `force` is specified by the user....")
                if e_pid:
                    raise RuntimeError("But pid exists so it's dangereous")

        # TODO: Problem with the default manager when creating the flow.
        # Replace node_id with uuid4
        config_dir = os.path.dirname(path)
        sched_options = yaml_safe_load_path(os.path.join(config_dir, "scheduler.yml"))
        manager = TaskManager.from_file(os.path.join(config_dir, "manager.yml"))

        # TODO: manager, mng_connector, flow_model_cls
        # Should be stored in state.json so that we can restart.
        new = cls(state.name, sched_options, state.scratch_dir, address=state.address, port=state.port,
                  manager=manager, mng_connector=state.mng_connector)

        print("Remember to execute lscan (rscan) to update the list of local (remote) clients...")
        return new

    def write_state_file(self, status: str = "dead", filepath: Optional[str] = None) -> str:
        """
        Write/update the state.json file of the worker.
        """
        if filepath is None:
            filepath = os.path.join(self.config_dir, "state.json")

        state = WorkerState(
            name=self.name,
            status=status,
            pid=self.pid,
            address=self.address,
            port=self.port,
            scratch_dir=self.scratch_dir,
            mng_connector=self.mng_connector,
        )

        state.json_write(filepath)
        return filepath

    #def remove_flows(self):
    #    """This requires locking the SQLite database."""

    def serve_worker_homepage(self):
        """
        Serve the landing page of the AbipyWorker website
        """
        import platform
        system, node, release, version, machine, processor = platform.uname()

        header = pn.pane.Markdown(f"""
# Homepage of **{self.name}** {self.__class__.__name__}

Running on {socket.gethostname()} -- system {system} -- Python {platform.python_version()}
""", sizing_mode="stretch_width")

        d = self.flow_scheduler.groupby_status()
        md_lines = ["Refresh the page to update the list of Abipy Flows."]
        if d:
            for flow_status, values in d.items():
                if flow_status == "error": flow_status = "Errored"
                md_lines.append(f"## {flow_status.capitalize()} Flows:\n")
                for row in values:
                    workdir, flow_id = row["workdir"], row["flow_id"]
                    md_lines.append(f"- [flow_id: {flow_id}, workdir: {workdir}](/flow/{flow_id})\n")
        else:
            md_lines.append("## Empty Flow list!")

        with open(os.path.join(self.config_dir, "state.json"), "rt") as fp:
            state = json.load(fp)

        from abipy.panels.viewers import JSONViewer
        main = pn.Column(
                header,
                pn.Row("## Server state", JSONViewer(state)),
                #pn.layout.Divider(),
                pn.pane.Markdown("\n".join(md_lines), sizing_mode="stretch_width"),
                sizing_mode="stretch_width",
        )

        from abipy.panels.core import get_abinit_template_cls_kwds
        tmpl_cls, tmpl_kwds = get_abinit_template_cls_kwds()
        return tmpl_cls(main=main, **tmpl_kwds)

    def serve_panel_flow(self):
        # URL example: /flow/123 where 123 is the flow id.
        tokens = pn.state.app_url.split("/")
        #print("In serve_panel_flow with app_url", pn.state.app_url, "tokens:", tokens)
        flow_id = int(tokens[2])
        flow, flow_status = self.flow_scheduler.get_flow_and_status_by_nodeid(flow_id)
        if flow is not None:
            return flow.get_panel(template=self.panel_template)

        return pn.pane.Alert(f"Cannot find Flow with node id: {flow_id}", alert_type="danger")

    def shutdown_server(self, message: str) -> None:
        """
        Shutdown the AbipyWorker.
        """
        self.write_state_file(status="dead")
        #self.flow_scheduler.stop()

        # If we are connected to a MongoDB collection, we try to log the problem in the database.
        #if self.flow_model_cls is not None:
        #    try:
        #        collection = self.mng_connector.get_collection()
        #        self.flow_model_cls.register_worker_exception(self, collection)
        #    except:
        #        pass

        print("Shuting the server down.")
        print(message)
        sys.exit(1)

    def serve(self, **serve_kwargs):
        """
        Start the webserver.
        """
        # Change the default manager returned by TaskManager.from_user_config.
        # so that we always use the manager in the AbipyWorker directory when costructing new Flows.
        assert self.manager is not None, "TaskManager should not be None"
        set_user_config_taskmanager(self.manager)

        # Register function atexit to update the state.json file if the worker gets killed.
        # Note that the callback won't be executed if we receive SIGKILL (kill -9)
        # or if an exception is raised in write_state_file or in any other registered callback.
        import atexit
        atexit.register(self.write_state_file)

        from abipy.panels.core import abipanel #, get_abinit_template_cls_kwds, AbipyParameterized
        self.panel_template = serve_kwargs.pop("panel_template", "FastList")
        abipanel(panel_template=self.panel_template)

        import abipy.panels as mod
        assets_path = os.path.join(os.path.dirname(mod.__file__), "assets")
        serve_kwargs.update(
            show=False,
            static_dirs={"/assets": assets_path},
            websocket_origin="*",
        )

        #print("serve_kwargs:", serve_kwargs)
        self.address = serve_kwargs.pop("address", self.address)
        self.port = serve_kwargs.pop("port", self.port)
        if self.port == 0:
            self.port = find_free_port(address=self.address)

        # Now write state.json with the actual port.
        self.write_state_file(status="serving")

        # Start the secondary thread for the MultiFlowScheduler.
        thread = threading.Thread(target=self.flow_scheduler.start, name="flow_scheduler", daemon=True)
        thread.start()
        #termcolor.enable(False)

        if self.flow_model_cls is not None:
            print("Abipy Worker will try to connect to the MongoDB server using:")
            print(self.mng_connector)
            print("Using FlowModel class:", self.flow_model_cls)
            print("Connecting every", self.MONGO_PERIODIC_CALLBACK_TIME_MS, "ms")
            from tornado.ioloop import PeriodicCallback
            periodic = PeriodicCallback(self.build_flow_from_mongodb, self.MONGO_PERIODIC_CALLBACK_TIME_MS)
            periodic.start()

        server = pn.serve(self.routes, port=self.port, address=self.address,
                          extra_patterns=self.extra_patterns, start=True, **serve_kwargs)

        #return server.start()


def _get_all_local_worker_states(dirpath: Optional[str] = None) -> list[WorkerState]:
    dirpath = ABIPY_DIRPATH if not dirpath else dirpath

    worker_dirs = [dirname for dirname in os.listdir(dirpath) if dirname.startswith("worker_")]
    states = []
    for workdir in worker_dirs:
        state = WorkerState.from_json_file(os.path.join(dirpath, workdir, "state.json"))
        states.append(state)

    return states


def print_local_workers(dirpath: Optional[str] = None) -> None:
    df = pd.DataFrame(_get_all_local_worker_states(dirpath=dirpath))
    print_dataframe(df, title="\nLocal AbiPy Workers:\n")



_SSH_CONFIG = None


def get_ssh_config():
    global _SSH_CONFIG

    if _SSH_CONFIG is None:
        # For the API, see http://docs.paramiko.org/en/stable/api/config.html#config-module-api-documentation
        from paramiko import SSHConfig
        user_config = os.path.expanduser("~/.ssh/config")
        if not os.path.exists(user_config):
            raise RuntimeError(f"Cannot find ssh configuration file at {user_config}")
        _SSH_CONFIG = SSHConfig.from_path(user_config)
        #print(_SSH_CONFIG)

    return _SSH_CONFIG

    # Return the set of literal hostnames defined in the SSH config
    # (both explicit hostnames and wildcard entries).
    #ssh_config.get_hostnames()
    #host_conf = ssh_config.lookup(self.hostname)


class ClientError(Exception):
    """Exceptions raised by WorkerClient"""

    @classmethod
    def from_response(cls, response):
        """Builds the object from a requests response object."""
        raise cls(f"\nServer returned status code: {response.status_code}\nwith text:\n{response.text}\n")


def need_serving_worker(method):
    """
    Decorator for methods of the WorkerClient that require a running AbipyWorker.
    """

    @functools.wraps(method)
    def decorated(*args, **kwargs):
        self: WorkerClient = args[0]
        if self.worker_state.status != self.worker_state.status.serving:
            raise ClientError(
                f"Status of {self.worker_state.name} worker is `{self.worker_state.status}`\n"
                f"Use `abiw.py restart {self.worker_state.name}` to restart the worker"
            )

        return method(*args, **kwargs)

    return decorated


class WorkerClient(MSONable):
    """
    The client used by the user to connect to an AbipyWorker, send python scripts or directories for execution.
    This object also provides an API to query the status of the Worker.
    """

    def __init__(self, worker_state, is_default_worker=False, timeout=240,
                 is_local_worker=None, ssh_destination=None, local_port=None, saved_uuid=None):

        if not isinstance(worker_state, WorkerState):
            self.worker_state = WorkerState(**worker_state)
        else:
            self.worker_state = worker_state

        self.is_default_worker = bool(is_default_worker)
        self.timeout = timeout

        if is_local_worker is None:
            # Well, this is not perfect as one may have a laptop with the same hostname as the remote cluster.
            self.is_local_worker = self.worker_state.hostname == socket.gethostname()
        else:
            self.is_local_worker = bool(is_local_worker)

        if ssh_destination is None:
            if self.is_local_worker:
                self.ssh_destination = ""
            else:
                hostnames = get_ssh_config().get_hostnames()
                if self.worker_state.hostname not in hostnames:
                    print(f"WARNING: Cannot find `{self.worker_state.hostname}` in `{hostnames}`")

                self.ssh_destination = f"{self.worker_state.username}@{self.worker_state.hostname}"
        else:
            self.ssh_destination = str(ssh_destination)

        if local_port is not None:
            self.local_port = int(local_port)
            self.server_url = f"http://localhost:{self.local_port}"

        else:
            if self.is_local_worker:
                self.local_port = self.worker_state.port
                if self.local_port is None:
                    raise ValueError("If local worker, local_port cannot be None")
                self.server_url = f"http://localhost:{self.local_port}"
            else:
                self.local_port = local_port
                if self.local_port is not None:
                    self.server_url = f"http://localhost:{self.local_port}"
                else:
                    self.server_url = None

        self.saved_uuid = saved_uuid if saved_uuid else self.worker_state.uuid

    @property
    def is_remote_worker(self) -> bool:
        """True if this is a remote worker."""
        return not self.is_local_worker

    #def lworker_dirpath(self) -> bool:
    #    dirname = "worker_{self.worker_state.name}"
    #    return os.path.expanduser(o.path.join("~", ".abinit". "abipy", dirname))

    @pmg_serialize
    def as_dict(self) -> dict:
        return dict(
                worker_state=self.worker_state.dict(),
                is_default_worker=self.is_default_worker,
                timeout=self.timeout,
                is_local_worker=self.is_local_worker,
                ssh_destination=self.ssh_destination,
                local_port=self.local_port,
                saved_uuid=self.saved_uuid,
        )

    #def from_dict(self, d):
    #    return self(**d)

    def setup_ssh_port_forwarding(self) -> None:
        # TODO Rationalize this part. Make sure hostname is listed in ~/.ssh/config.
        info = []

        # Return immediately if port forwarding is not needed.
        if self.is_local_worker or self.worker_state.status != self.worker_state.status.serving:
            return

        print("in setup_ssh_port_forwarding")
        # Need ssh port-forwarding.
        if self.local_port is None:
            self.local_port = find_free_port()
        else:
            if not port_is_open(self.local_port):
                self.local_port = find_free_port()
            else:
                info.append("port self.local_port is already opened")
                return

        ssh_cmd = f"ssh -N -f -L localhost:{self.local_port}:localhost:{self.worker_state.port} {self.ssh_destination}"
        print(f"executing: `{ssh_cmd}`")
        subprocess.run(ssh_cmd, shell=True)

        #return info

    def update_state(self, worker_state: WorkerState) -> None:
        """
        Update the state of the Worker stored by the client.
        """
        prev_hostname = self.worker_state.hostname
        self.worker_state = worker_state.copy()
        # This is needed to preserve the changes done by the user in clients.json
        # if the hostname differs from the one used in the ~/.ssh/config file.
        self.worker_state.hostname = prev_hostname

        if self.saved_uuid != worker_state.uuid:
            print("Detected change in worker_state.uuid!")
            self.local_port = None
            if self.is_default_worker:
                self.local_port = worker_state.port

            self.saved_uuid = worker_state.uuid

    def __repr__(self) -> str:
        return repr(self.worker_state)

    def __str__(self) -> str:
        return self.to_json()

    def to_json(self) -> str:
        """
        Returns a string with the json representation of the client
        """
        # TODO: monty.to_json should accept **kwargs
        return json.dumps(self, cls=AbipyEncoder, indent=2)

    def check_server_url(self) -> None:
        if self.server_url is None:
            raise ClientError("server_url cannot be None.")
        if self.local_port is None:
            raise ClientError("localport cannot be None.")
        if not port_is_open(self.local_port):
            raise ClientError("local port {local_port} is not open.")

    def _get_url(self, end_point: str, action: str) -> str:
        url = f"{self.server_url}/{end_point}"
        print(f"{action} to {self.worker_state.name} worker at url: {url} ...\n")
        self.check_server_url()
        return url

    ##################################################
    # HTTP interface implemented via requests package
    ##################################################

    @need_serving_worker
    def send_pyscript(self, filepath, user_message="", end_point="post_flow_script"):
        """
        Send a python script `filepath` to the AbipyWorker in order to build a new Flow.
        Returns the json-encoded content of the response, if any.
        """
        filepath = os.path.expanduser(filepath)
        with open(filepath, "rt") as fp:
            data = dict(
                pyscript_basename=os.path.basename(filepath),
                pyscript_string=fp.read(),
                user_message=user_message,
            )

        url = self._get_url(end_point, f"Sending `{filepath}` script")
        r = requests.post(url=url, json=data, timeout=self.timeout)
        if r.status_code == 200:
            return r.json()

        raise ClientError.from_response(r)

    @need_serving_worker
    def send_flow_dir(self, flow_dir: str, user_message: str = "", end_point="post_flow_dirs"):
        """
        Send a directory containing a Flow to the AbipyWorker.
        Returns the json-encoded content of the response, if any.

        Args:
            flow_dir: Directory with the Flow (already built by the user)
            user_message:
            end_point:
        """
        if not self.is_local_worker:
            raise RuntimeError(f"You cannot add a Flow to the remote worker: {self.worker_state.name}")

        if not os.path.isdir(flow_dir):
            raise RuntimeError(f"Cannot find flow directory: {flow_dir}")

        url = self._get_url(end_point, f"Sending `{flow_dir}` directory")
        data = dict(flow_dir=flow_dir, user_message=user_message)

        r = requests.post(url=url, json=data, timeout=self.timeout)
        if r.status_code == 200:
            return r.json()

        raise ClientError.from_response(r)

    @need_serving_worker
    def send_kill_message(self, end_point="action") -> str:
        """
        Ask the server to terminate execution.
        """
        url = self._get_url(end_point, "Sending kill message")
        r = requests.post(url=url, json=dict(action="kill"), timeout=self.timeout)
        if r.status_code == 200:
            return r.json()

        raise ClientError.from_response(r)

    @need_serving_worker
    def get_json_status(self, end_point="json_status") -> str:
        """
        Inquire server status.
        """
        url = self._get_url(end_point, "Sending status request")
        r = requests.get(url=url, timeout=self.timeout)

        if r.status_code == 200:
            return r.json()
            #print(r.text)
            #json_status = r.json()
            #from pandas.io.json import read_json
            #json_status["dataframe"] = read_json(json_status["dataframe"])
            #print_dataframe(json_status["dataframe"], title="\nWorker Status:\n")
            #return json_status

        raise ClientError.from_response(r)

    @need_serving_worker
    def open_webgui(self):
        """
        Open a new tab in the browser with a panel dashboard to interact with the Worker.
        """
        end_point = ""
        url = self._get_url(end_point, f"Contacting {self.worker_state.name} worker")
        #print(self)
        import webbrowser
        webbrowser.open_new_tab(url)


class WorkerClients(list, MSONable):
    """
    A list of Clients.
    """

    @classmethod
    def lscan(cls, dirpath: Optional[str] = None) -> WorkerClients:
        """
        Discover all the AbipyWorkers created on the same machine by inspecting
        the configuration directories in the AbiPy directory.

        Return: list of clients associated to local AbiPy workers.
        """
        worker_states = _get_all_local_worker_states(dirpath)
        all_clients = cls.from_json_file(empty_if_not_file=True)
        all_clients = all_clients.remove_stale_local_clients()
        all_clients.update_worker_states(worker_states, write_json=True)

        return cls(c for c in all_clients if c.is_local_worker)

    @classmethod
    def from_json_file(cls, empty_if_not_file: bool = False, filepath: Optional[str] = None) -> WorkerClients:
        """
        Build object from JSON file.

        Args:
            empty_if_not_file:
            filepath:
        """
        filepath = os.path.expanduser(filepath) if filepath else os.path.join(ABIPY_DIRPATH, "clients.json")

        new = cls()
        if empty_if_not_file and not os.path.exists(filepath):
            return new

        if not os.path.exists(filepath):
            raise RuntimeError(f"""
Could not find `{filepath}`
Please execute:

    abiw.py lscan

or

    abiw.py rscan

to create it""")

        with open(filepath, "rt") as fp:
            data = json.load(fp)
            for d in data["clients"]:
                client = WorkerClient.from_dict(d)
                new.append(client)

        new._validate()
        return new

    @classmethod
    def rscan(cls, hostnames: list[str], printout: bool = True) -> WorkerClients:
        """
        Uses fabric to discover AbipyWorkers running on remote machines.

        Args:
            hostnames: List of hostnames as declared in ~/.ssh/config.
            printout:
        """
        #
        # From https://docs.fabfile.org/en/2.6/api/transfer.html#fabric.transfer.Transfer
        #
        # Most SFTP servers set the remote working directory to the connecting user’s home directory,
        # and (unlike most shells) do not expand tildes (~).
        #
        # For example, instead of saying get("~/tmp/archive.tgz"), say get("tmp/archive.tgz").

        try:
            from fabric import Connection
        except ImportError:
            raise ImportError("rscan requires fabric package. Install it with `pip install fabric`.")

        from io import BytesIO
        print("Make sure the hosts are listed in your ~/.ssh/config file and that ProxyJump is activated if needed")
        worker_states = []
        for host in tqdm(hostnames):
            print(f"Contacting {host}. It may take some time ...")
            try:
                c = Connection(host)
                result = c.run("ls ~/.abinit/abipy")
                files = result.stdout.split()
                worker_dirs = [f for f in files if f.startswith("worker_")]
                worker_dirs = [os.path.join(".abinit/abipy", d) for d in worker_dirs]
            except Exception as exc:
                print(exc)
                continue

            # Now get the state.json via sftp.
            for w in worker_dirs:
                state_path = os.path.join(w, "state.json")
                try:
                    strio = BytesIO()
                    c.get(state_path, local=strio)
                    json_bstring = strio.getvalue()
                    strio.close()
                    worker_states.append(WorkerState.from_json(json_bstring))

                except IOError as exc:
                    print(f"Cannot find `state.json` file: {host}@{state_path}. Ignoring exception.")
                    print(exc)

        all_clients = cls.from_json_file(empty_if_not_file=True)
        all_clients = all_clients.remove_stale_local_clients()
        all_clients.update_worker_states(worker_states, write_json=False)

        #for client in all_clients:
        #    client.setup_ssh_port_forwarding()

        all_clients.write_json_file()

        return cls(c for c in all_clients if c.is_remote_worker)

    def _validate(self) -> None:
        err_lines = []
        app = err_lines.append
        from collections import Counter

        # worker_name must be unique.
        for worker_name, count in Counter((w.worker_state.name for w in self)).items():
            if count > 1:
                app(f"Server name: `{worker_name}` appears: {count} times. This is forbidden!")

        # Only zero or one default server is allowed.
        count = sum(1 if w.is_default_worker == True else 0 for w in self)
        if count not in (0, 1):
            app(f"is_default_worker=True appears `{count}` times." +
                "This is forbidden as only one default worker is allowed!")

        #self.check_local_port()

        for client in self:
            #print(client.local_port, "\n", client)
            if duck.is_intlike(client.local_port):
                address = "localhost"
                if not port_is_open(client.local_port, address=address):
                    #print(f"WARNING: Nobody is listening to `{address}:{client.local_port}`")
                    client.local_port = None
                else:
                    print(f"{client.worker_state.name} listening to `{address}:{client.local_port}`")

        if err_lines:
            raise RuntimeError("\n".join(err_lines))

    def __str__(self):
        lines = [str(w) for w in self]
        return (2 * "\n").join(lines)

    #@pmg_serialize
    def as_dict(self) -> dict:
        return {"clients": [w.as_dict() for w in self]}

    #def dict(self) -> dict:
    #    return {"clients": [w.dict() for w in self]}

    def write_json_file(self, filepath: Optional[str] = None) -> None:
        """
        Write json file with the list of clients
        """
        filepath = os.path.join(ABIPY_DIRPATH, "clients.json") if not filepath else os.path.expanduser(filepath)

        # TODO: Compare old with new. return diff
        #old_clients = self.__class__.from_json_file(empty_if_not_file=True)
        #if old_clients is None:

        with open(filepath, "wt") as fp:
            json.dump(self.as_dict(), fp, indent=2)

    def update_worker_states(self, worker_states: list[WorkerState], write_json: bool = True) -> None:
        """
        Update clients from worker_states
        """
        for state in worker_states:
            client = self.select_from_worker_name(state.name, allow_none=True)
            if client is not None:
                client.update_state(state)
            else:
                self.append(WorkerClient(state))

        self._validate()
        if write_json: self.write_json_file()

    def remove_stale_local_clients(self) -> WorkerClients:
        """
        Return new list of clients where all entries associated to workers whose
        directory is not found have been removed.
        """
        new_clients = []
        for client in self:
            worker_dirpath = os.path.join(ABIPY_DIRPATH, f"worker_{client.worker_state.name}")
            if client.is_local_worker and not os.path.exists(worker_dirpath):
                print("Removing stale section with worker:", client.worker_state.name)
            else:
                new_clients.append(client)

        return self.__class__(new_clients)

    def get_dataframe(self) -> pd.DataFrame:
        """
        Build and return pandas Dataframe with the AbipyWorker states.
        """
        rows = []
        for client in self:
            d = client.as_dict()
            d = {k: v for k, v in d.items() if not k.startswith("@")}
            state = d.pop("worker_state", )
            if "status" in state:
                state["status"] = state["status"].value  #.colored_str()

            mng_connector_dict = state.pop("mng_connector", None)
            if mng_connector_dict is not None:
                # Add the most important info from the mng_connector.
                # Prepend some keys with `mongo_prefix`.
                remap = {"host": "mongo_host", "dbname": "mongo_dbname", "collection_name": None}
                mongo_info = {k: mng_connector_dict[k] for k in remap}
                for k, new_key in remap.items():
                    if new_key is None: continue
                    mongo_info[new_key] = mongo_info.pop(k)

                state.update(**mongo_info)

            d.update(**state)
            rows.append(d)

        df = pd.DataFrame(rows)
        return df

    def print_dataframe(self) -> None:
        #if select_keys == "short":
        #    selected_keys = ["name", "status", "pid", "scratch_dir", "username", "hostname", ]

        #all_keys = ["is_default_worker", "timeout", "is_local_worker", "ssh_destination",
        #            "local_port", "name",  "status", "pid", "address",  "port",  scratch_dir",
        #            "username", "hostname" "uuid version"
        #        ]

        df = self.get_dataframe().set_index("name")
        print_dataframe(df, title="\nAbiPyWorker Clients:\n")

    def summarize(self) -> pd.DataFrame:
        """
        Return dataframe with the most important columns.
        """
        select_keys = ["name", "status", "pid", "scratch_dir"]
        if any(c.is_remote_worker for c in self):
            select_keys += ["username", "hostname"]
        df = self.get_dataframe()

        if df.empty: return df
        return df[select_keys].set_index("name")

    def get_all_worker_names(self) -> list[str]:
        """Return the list of worker names stored in self."""
        return [client.worker_state.name for client in self]

    def select_from_worker_name(self, worker_name: str,
                                allow_none=False) -> Union[WorkerClient, None]:
        """
        Select and return a WorkerClient from its name `worker_name`.

        Return: None if `worker_name` is not found and `allow_none` is True else raises ValueError.
        """
        if worker_name is None:
            # If there are n > 1 workers, a default must be set by the user.
            for client in self:
                if client.is_default_worker: return client

            if len(self) == 1: return self[0]

            raise ValueError("Cannot find default worker in list:\n" +
                             f"\t{self.get_all_worker_names()}\n" +
                             "Use:\n\tabiw.py set_default WORKER_NAME")

        for client in self:
            if client.worker_state.name == worker_name: return client

        if allow_none: return None
        raise ValueError(f"Cannot find worker with name `{worker_name}`. Choose among:\n" +
                         f"\t{self.get_all_worker_names()}\n")

    def set_default(self, worker_name: str) -> None:
        """
        Set the Client with `worker_name` as the default one.
        """
        the_one = self.select_from_worker_name(worker_name, allow_none=False)
        for client in self:
            client.is_default_worker = False
        the_one.is_default_worker = True

        self._validate()
        self.write_json_file()





import subprocess
import random
import tempfile

class SSHTunnel:
    """
    https://stackoverflow.com/questions/4364355/how-to-open-an-ssh-tunnel-using-python

    And here's how you could use it, for example with MySQL (port 3306 usually):

    with SSHTunnel('database.server.com', 'you', '22', '/path/to/private_key', '3306') as tunnel:
        print("Connected on port {} at {}".format(tunnel.local_port, tunnel.local_host))
    """

    def __init__(self, host, user, port, key, remote_port):
        self.host = host
        self.user = user
        self.port = port
        self.key = key
        self.remote_port = remote_port
        # Get a temporary file name
        tmpfile = tempfile.NamedTemporaryFile()
        tmpfile.close()
        self.socket = tmpfile.name
        self.local_port = random.randint(10000, 65535)
        self.local_host = '127.0.0.1'
        self.open = False

    def start(self):
        # I had to use -o ExitOnForwardFailure=yes instead of -o ExitOnForwardFailure=True.
        exit_status = subprocess.call(['ssh', '-MfN',
            '-S', self.socket,
            '-i', self.key,
            '-p', self.port,
            '-l', self.user,
            '-L', '{}:{}:{}'.format(self.local_port, self.local_host, self.remote_port),
            '-o', 'ExitOnForwardFailure=yes',
            self.host
        ])
        if exit_status != 0:
            raise Exception('SSH tunnel failed with status: {}'.format(exit_status))
        if self.send_control_command('check') != 0:
            raise Exception('SSH tunnel failed to check')
        self.open = True

    def stop(self):
        if self.open:
            if self.send_control_command('exit') != 0:
                raise Exception('SSH tunnel failed to exit')
            self.open = False

    def send_control_command(self, cmd):
        return subprocess.check_call(['ssh', '-S', self.socket, '-O', cmd, '-l', self.user, self.host])

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()







# coding: utf-8
"""
Object to analyze the results stored in the WR.nc file
"""
import numpy as np

from monty.string import marquee
from monty.functools import lazy_property
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt #, get_axarray_fig_plt
from abipy.core.mixins import AbinitNcFile, Has_Structure, NotebookWriter
from abipy.iotools import ETSF_Reader


class WrNcFile(AbinitNcFile, Has_Structure, NotebookWriter):

    def __init__(self, filepath):
        super().__init__(filepath)
        self.reader = r = ETSF_Reader(filepath)

        # Read dimensions.
        self.nfft = r.read_dimvalue("nfft")
        self.nspden = r.read_dimvalue("nspden")
        self.natom3 = len(self.structure) * 3
        self.method = r.read_value("method")
        assert self.method == 0
        self.ngqpt = r.read_value("ngqpt")
        self.rpt = r.read_value("rpt")
        self.nrpt = len(self.rpt)
        # FFT mesh.
        self.ngfft = r.read_value("ngfft")

    def create_xsf(self, iatom=0, red_dir=(1, 0, 0), u=1.0, ispden=0):

        nfft, nrpt = self.nfft, self.nrpt

        nx, ny, nz = self.ngfft
        nqx, nqy, nqz = self.ngqpt
        box_shape = self.ngqpt * self.ngfft
        box_size = np.product(box_shape)
        print("ngqpt:", self.ngqpt)
        print("nrpt:", self.nrpt)
        print("Unit cell FFT shape:", self.ngfft)
        print("Big box shape:", box_shape)
        print("rpt:\n", self.rpt)

        # Get FFT points in reduced coordinates of the microcell.
        # ix is the fastest index here because we are gonna access
        # FFT values produced by Fortran via ifft
        fft_inds = np.empty((nfft, 3), dtype=np.int64)
        ifft = -1
        for iz in range(nz):
            for iy in range(ny):
                for ix in range(nx):
                    ifft += 1
                    fft_inds[ifft, :] = [ix, iy, iz]

        def ig2gfft(ig, ng):
            # Use the following indexing (N means ngfft of the adequate direction)
            # 0 1 2 3 ... N/2    -(N-1)/2 ... -1    <= gc
            # 1 2 3 4 ....N/2+1  N/2+2    ...  N    <= index ig

            #if ( ig <= 0 or ig > ng):
            #  # Wrong ig, returns huge. Parent code will likely crash with SIGSEV.
            #  gc = huge(1)
            #  return

            #if (ig  > ng/2 + 1):
            #  gc = ig - ng -1
            #else
            #  gc = ig -1
            #return gc
            raise NotImplementedError("Foo")

        # Find index of (r, R) in the bigbox
        # Use C notation (z runs faster)
        #box2fr = np.full((box_size, 2), np.inf, dtype=np.int64)

        #iffr2box = np.empty((nfft, nrpt), dtype=np.int64)
        print("Building r-R dictionary")
        d = {}
        for ir, rpt in enumerate(self.rpt):
            for ifft, fft_ijk in enumerate(fft_inds):
                ijk = (fft_ijk - rpt * self.ngfft) % box_shape
                key = tuple(map(int, ijk))
                d[key] = (ifft, ir)
                #i, j, k = ijk
                #box_iloc = k + j * box_shape[2] + i * (box_shape[2] * box_shape[1])
                #print(box_iloc, "(i, j, k): ", fft_ijk, "rpt:", rpt)
                #box2fr[int(box_iloc)] = [ifft, ir]
                #iffr2box[ifft, ir] = box_iloc
        print("Done")

        #ip = 0
        #idir = ip % 3
        #iatom = (ip - idir) // 3 # + 1

        # nctkarr_t("v1scf_rpt_sr", "dp", "two, nrpt, nfft, nspden, natom3")
        # use iorder = "f" to transpose the last 3 dimensions since ETSF
        # stores data in Fortran order while AbiPy uses C-ordering.
        # (z,y,x) --> (x,y,z)
        # datar = transpose_last3dims(datar)
        wsr_var = self.reader.read_variable("v1scf_rpt_sr")
        wlr_var = self.reader.read_variable("v1scf_rpt_lr")

        wsr = np.zeros(nfft, nrpt)
        wlr = np.zeros(nfft, nrpt)
        for idir, red_comp in enumerate(red_dir):
            ip = idir + 3 * iatom
            wsr += u * red_comp * wsr_var[ip, ispden, :, :, 0]
            wlr += u * red_comp * wlr_var[ip, ispden, :, :, 0]

        print("wsr.shape:", wsr.shape)
        print("Max |Re Wsr|:", np.max(np.abs(wsr.real)), "Max |Im Wsr|:", np.max(np.abs(wsr.imag)))
        #print("Max |Re Wlr|:", np.max(np.abs(wlr.real)), "Max |Im Wlr|:", np.max(np.abs(wlr.imag)))

        r0 = np.array([0, 0, 0], dtype=np.int)
        qgrid = np.where(self.ngqpt > 2, self.ngqpt, 0)
        r0 = - (self.ngqpt - 1) // 2
        print("Origin of datagrid set at R0:", r0)

        # Build datagrid in the supercell using C indexing
        # This is what xsf_write_data expects.
        miss = []
        data_lr = np.empty(box_shape)
        data_sr = np.empty(box_shape)

        print("Filling data array")
        for ix in range(box_shape[0]):
            for iy in range(box_shape[1]):
                for iz in range(box_shape[2]):
                    y = np.array((ix, iy, iz), dtype=np.int)
                    x = (y + r0) % box_shape
                    key = tuple(map(int, x))
                    try:
                        ifft, ir = d[key]
                    except KeyError:
                        #print("Cannot find r - R with key:", key)
                        miss.append(key)
                        continue

                    data_lr[ix, iy, iz] = wlr[ifft, ir]
                    #data_sr[ix, iy, iz] = wsr[ifft, ir]

        if miss:
            #print(d.keys())
            raise RuntimeError("Cannot find r-R points! nmiss:", len(miss))

        super_structure = self.structure * self.ngqpt

        def dump_xsf(filename, data):
            from abipy.iotools import xsf
            xsf.xsf_write_structure_and_data_to_path(filename, super_structure, data, cplx_mode="abs")

        dump_xsf("foo_lr.xsf", data_lr)
        dump_xsf("foo_sr.xsf", data_sr)

    @lazy_property
    def structure(self):
        """|Structure| object."""
        return self.reader.read_structure()

    def close(self):
        self.reader.close()

    @lazy_property
    def params(self):
        """:class:`OrderedDict` with parameters that might be subject to convergence studies."""
        return {}

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose=0):
        """String representation."""
        lines = []; app = lines.append
        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")

        return "\n".join(lines)

    @add_fig_kwargs
    def plot_maxw(self, scale="semilogy", ax=None, fontsize=8, **kwargs):
        """
        Plot the decay of max_{r,idir,ipert} `|W(R,r,idir,ipert)|`
        for the long-range and the short-range part.

        Args:
            scale: "semilogy", "loglog" or "plot".
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: fontsize for legends and titles

        Return: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        f = {"plot": ax.plot, "semilogy": ax.semilogy, "loglog": ax.loglog}[scale]

        rmod = self.reader.read_value("rmod")

        # Plot short-range part.
        # normalize wrt the R=0 value
        # Fortran array: nctkarr_t("maxw_sr", "dp", "nrpt, natom3")
        maxw_sr = self.reader.read_value("maxw_sr")
        data = np.max(maxw_sr, axis=0)
        #data = data / data[0]
        f(rmod, data, marker="o", ls=":", lw=0, label="SR", **kwargs)

        # Plot long-range part.
        maxw_lr = self.reader.read_value("maxw_lr")
        data = np.max(maxw_lr, axis=0)
        #data = data / data[0]
        f(rmod, data, marker="x", ls="-", lw=0, label="LR", **kwargs)

        # Plot the ratio
        data = np.max(maxw_lr, axis=0) / np.max(maxw_sr, axis=0)
        #f(rmod, data, marker="x", ls="-", lw=0, label="LR/SR", **kwargs)

        #rmod = self.reader.read_value("rmod_lrmodel")
        #maxw = self.reader.read_value("maxw_lrmodel")
        #data = np.max(maxw, axis=0)
        #data = data / data[0]
        #f(rmod, data, marker="x", ls="-", lw=0, label="W_LR_only", **kwargs)

        ax.grid(True)
        ax.set_ylabel(r"$Max_{({\bf{r}}, idir, ipert)} \| W({\bf{r}}, {\bf{R}}, idir, ipert) \|$")
        ax.set_xlabel(r"$\|{\bf{R}}\|$ (Bohr)")
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        #if kwargs.pop("with_title", True):
        #    ax.set_title("dvdb_add_lr %d, qdamp: %s, symv1scf: %d" % (self.dvdb_add_lr, self.qdamp, self.symv1scf),
        #                 fontsize=fontsize)
        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot_maxw(scale="semilogy")

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ncfile)"),
        ])

        #nb.cells.append(nbv.new_code_cell("ncfile.plot_diff_at_qpoint(qpoint=%d);" % iq))

        return self._write_nb_nbpath(nb, nbpath)


if __name__ == "__main__":
    import sys
    ncfile = WrNcFile.from_file(sys.argv[1])

    #print(ncfile)
    ncfile.plot_maxw(scale="semilogy", ax=None, fontsize=8)
    #ncfile.create_xsf(iatom=0, red_dict=(-1, +1, +1), u=0.1)



# coding: utf-8
"""Work subclasses for the computation of luminiscent properties."""
from __future__ import annotations

from .works import Work, PhononWork


class LumiWork(Work):
    """
    This Work implements Fig 1 of https://arxiv.org/abs/2010.00423.

    Client code is responsible for the preparation of the supercell and
    of the GS SCF input files with the fixed electronic occupations associated to the two configurations.
    By default, the work computes four total energies corresponding to the Ag, Ag*, Ae*, Ae configurations.
    Optionally, one can activate the computation of four electronic band structures
    and phonons for Ag and Ae. See docstring of from_scf_inputs for further info.
    """

    @classmethod
    def from_scf_inputs(cls, gs_scf_inp, exc_scf_inp, relax_kwargs, ndivsm=0, nb_extra=10,
                        tolwfr=1e-20, ngqpt=None, manager=None) -> LumiWork:
        """
        Args:
            gs_scf_inp: |AbinitInput| representing a GS SCF run for the ground-state.
            exc_scf_inp: |AbinitInput| representing a GS SCF run for the excited-state.
            relax_kwargs: Dictonary with input variables to be added to gs_scf_inp and exc_scf_inp.
                when generating input files for structural relaxations.
            ndivsm: Activates the computation of band structure if different from zero.
                if > 0, it's the number of divisions for the smallest segment of the path (Abinit variable).
                if < 0, it's interpreted as the pymatgen `line_density` parameter in which the number of points
                in the segment is proportional to its length. Typical value: -20.
                This option is the recommended one if the k-path contains two high symmetry k-points that are very close
                as ndivsm > 0 may produce a very large number of wavevectors.
            nb_extra: Number of extra bands added to the input nband when computing band structures (ndivsm != 0).
            tolwfr: Tolerance on the residuals used for the NSCF band structure calculations.
            ngqpt: If not None, activates computation of phonons with DFPT using this q-mesh.
                Usually Gamma-only that is ngqpt = [1, 1, 1].
            manager: |TaskManager| of the work. If None, the manager is initialized from the config file.
        """
        new = cls(manager=manager)

        # Templates for GS SCF calculations.
        new.gs_scf_inp = gs_scf_inp
        new.exc_scf_inp = exc_scf_inp

        # Save paramenters for the generation of input files at runtime.
        new.relax_kwargs = relax_kwargs
        new.ndivsm = int(ndivsm)
        new.tolwfr = tolwfr
        new.nb_extra = int(nb_extra)
        new.ngqpt = ngqpt

        # Relaxation for the Ag configuration.
        new.gs_relax_task = new.register_relax_task(gs_scf_inp.new_with_vars(relax_kwargs))

        if new.ngqpt is not None:
            # Make sure we have the WFK file if we are gonna compute phonons.
            new.gs_relax_task.input["prtwf"] = 1

        # Internal counter used in on_all_ok to drive the differ steps of the calculations.
        new.iteration_step = 0

        # JSON-compatible dictionary storing the most important results of the Work.
        # Results will be stored in the `lumi.json` file in the outdata directory of the Work
        # so that one can easily implement additional post-processing tools.
        new.json_data = {}

        return new

    def on_all_ok(self):
        """
        This method is called when all the tasks in the work have reached S_OK.

        This is the section in which we implement most of the workflow logic at runtime.
        since we need to generate input files with relaxed structures.
        """

        if self.iteration_step == 0:
            print("in iteration step 0")
            self.iteration_step += 1

            # Get Ag relaxed structure and energy.
            with self.gs_relax_task.open_gsr() as gsr:
                ag_relaxed_structure = gsr.structure
                self.json_data["Ag_energy_ev"] = float(gsr.energy)
                self.json_data["Ag_gsr_filepath"] = gsr.filepath

            # Build GS SCF input for the Ag* configuration:
            # use same structure as Ag but with excited occupation factors.
            exc_scf_inp = self.exc_scf_inp.new_with_structure(ag_relaxed_structure)
            self.agstar_scf_task = self.register_scf_task(exc_scf_inp)

            if self.ndivsm != 0:
                # Compute band structure for Ag* configuration.
                self.agstar_scf_task.add_ebands_task_to_work(self, ndivsm=self.ndivsm,
                                                             tolwfr=self.tolwfr, nb_extra=self.nb_extra)

            #if self.ngqpt is not None:
            #    Create new work to compute phonons for Ag configuration.
            #    TODO: Make sure we can connect DFPT with RelaxTask.
            #    self.ae_scf_task["prtwf"] = 1
            #    ph_work = PhononWork.from_scf_task(scf_task, self.ngqpt, is_ngqpt=True, tolerance=None, with_becs=True,
            #                                       ddk_tolerance=None, manager=None)
            #    self.flow.register_work(ph_work)
            #    Save location of final DDB file in json_data.
            #    self.json_data["Ag_ddb_filepath"] = ph_work.outdir.path_in("out_DDB")

            # Relax geometry with excited configuration starting from Ag*.
            self.agstar_relax_task = self.register_relax_task(exc_scf_inp.new_with_vars(self.relax_kwargs))

            return self.postpone_on_all_ok()

        elif self.iteration_step == 1:
            print("in iteration step 1")
            self.iteration_step += 1

            # Get Ag* total energy.
            with self.agstar_scf_task.open_gsr() as gsr:
                self.json_data["Agstar_energy_ev"] = float(gsr.energy)
                self.json_data["Agstar_gsr_filepath"] = gsr.filepath

            # Get Ae* relaxed structure and total energy.
            with self.agstar_relax_task.open_gsr() as gsr:
                aestar_relaxed_structure = gsr.structure
                self.json_data["Aestar_energy_ev"] = float(gsr.energy)
                self.json_data["Aestar_gsr_filepath"] = gsr.filepath

            # Build GS SCF input for the Ae configuration:
            # same structure as Ae* but with gs occupation factors.
            ae_scf_inp = self.gs_scf_inp.new_with_structure(aestar_relaxed_structure)
            self.ae_scf_task = self.register_scf_task(ae_scf_inp)

            if self.ndivsm != 0:
                # Compute band structure for Ae configuration.
                self.ae_scf_task.add_ebands_task_to_work(self, ndivsm=self.ndivsm,
                                                         tolwfr=self.tolwfr, nb_extra=self.nb_extra)

            if self.ngqpt is not None:
                # Create new work to compute phonons for the Ae configuration. Make sure prtwf == 1
                self.ae_scf_task.input["prtwf"] = 1
                ph_work = PhononWork.from_scf_task(self.ae_scf_task, self.ngqpt, is_ngqpt=True,
                                                   tolerance=None, with_becs=True, ddk_tolerance=None)
                self.flow.register_work(ph_work)
                # Save location of final DDB file in json_data.
                self.json_data["Ae_ddb_filepath"] = ph_work.outdir.path_in("out_DDB")

            return self.postpone_on_all_ok()

        elif self.iteration_step == 2:
            print("in iteration step 2")
            self.iteration_step += 1

            # Get Ag* total energy.
            with self.ae_scf_task.open_gsr() as gsr:
                ae_structure = gsr.structure
                self.json_data["Ae_energy_ev"] = float(gsr.energy)
                self.json_data["Ae_gsr_filepath"] = gsr.filepath

            # Write json file in the outdire of the work
            self.write_json_in_outdir("lumi.json", self.json_data)

            return super().on_all_ok()



#!/usr/bin/env python
r"""
Optic results
=============

This example shows how to plot the optical properties
computed by optic within the independent-particle approximation,
no local-field effects and no excitonic effects.
"""
from abipy import abilab
import abipy.data as abidata

# Here we use one of the OPTIC.nc files shipped with abipy.
# Replace filename with the path to your OPTIC.nc file.
filename = abidata.ref_file("gaas_121212_OPTIC.nc")

ncfile = abilab.abiopen(filename)

# Optic files have a Structure and an ElectronBands object.
#ncfile.ebands.plot()

# To plot linear dielectric tensor and other optical
# properties for all tensor components available in the file:
ncfile.plot_linopt(title="Linear dielectric function")

# To plot the second Harmonic tensor
ncfile.plot_shg(title="Second Harmonic Generation")

# Remember to close the file.
ncfile.close()

# Use OpticRobot to analyze multiple file e.g. convergence studies.
filenames = [
    abidata.ref_file("gaas_444_OPTIC.nc"),
    abidata.ref_file("gaas_888_OPTIC.nc"),
    abidata.ref_file("gaas_121212_OPTIC.nc"),
]

robot = abilab.OpticRobot.from_files(filenames)

# sphinx_gallery_thumbnail_number = 3
robot.plot_linopt_convergence(title=r"Convergence of $\epsilon$ wrt nkpt")

robot.plot_shg_convergence(title=r"Convergence of $\chi^2(-2\omega,\omega,\omega)$ wrt nkpt")

# Remember to close the file or use the `with` context manager.
robot.close()



#!/usr/bin/env python
r"""
e-ph ZPR at the band edges with the generalized Frohlich model
==============================================================

Flow to estimate the zero-point renormalization at the band edges
using the generalized Frohlich model. The flow uses DFPT to compute
the effective masses at the band edges (automatically detected by performing a NSCF run with a k-path),
BECS, eps_inf and phonon frequencies at Gamma
"""

import sys
import os
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def make_scf_input(usepaw=0):
    """Returns the GS input file"""
    # Here we use parameters similar to https://docs.abinit.org/tests/v8/Input/t57.in
    pseudos = abidata.pseudos("Ca.psp8", "O.psp8")

    structure = dict(
        acell=3 * [9.136],
        xred=[
           0.0000000000, 0.0000000000, 0.0000000000,
           0.5000000000, 0.5000000000, 0.5000000000],
        rprim=[
           0  , 0.5, 0.5,
           0.5, 0  , 0.5,
           0.5, 0.5, 0],
        typat=[1, 2],
        natom=2,
        ntypat=2,
        znucl=[20, 8],
    )

    scf_input = abilab.AbinitInput(structure=structure, pseudos=pseudos)

    scf_input.set_vars(
        nband=12,
        nbdbuf=2,
        diemac=6,
        ecut=30,               # Underconverged ecut.
        #ecut=15,
        nstep=100,
        tolvrs=1e-16,
        kptrlatt=[-2,  2,  2,  # In cartesian coordinates, this grid is simple cubic
                   2, -2,  2,
                   2,  2, -2],
    )

    return scf_input


def build_flow(options):
    # Set working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Build the SCF input.
    scf_input = make_scf_input()

    # Build the flow.
    from abipy.flowtk.effmass_works import FrohlichZPRFlow
    flow = FrohlichZPRFlow.from_scf_input(options.workdir, scf_input, ndivsm=4, tolwfr=1e-16,
                                          manager=options.manager)

    return flow


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_frohlich_zpr -s
#



#!/usr/bin/env python
"""
This benchmark compares GS calculations at the Gamma point done with istwfk in [1,2]
"""
import sys
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from abipy.benchmarks import bench_main, BenchmarkFlow


def make_input(paw=False):
    """
    Build and return an input file for GS calculations with paral_kgb=1
    """
    pseudos = abidata.pseudos("14si.pspnc") if not paw else abidata.pseudos("Si.GGA_PBE-JTH-paw.xml")
    structure = abidata.structure_from_ucell("Si")

    inp = abilab.AbinitInput(structure, pseudos)
    inp.set_kmesh(ngkpt=[1,1,1], shiftk=[0,0,0])

    # Global variables
    ecut = 20
    inp.set_vars(
        ecut=ecut,
        pawecutdg=ecut*4 if paw else None,
        nsppol=1,
        nband=20,
        paral_kgb=1,
        istwfk="*1",
        timopt=-1,
        chksymbreak=0,
        prtwf=0,
        prtden=0,
        tolvrs=1e-8,
        nstep=50,
    )

    return inp


def build_flow(options):
    template = make_input()

    # Get the list of possible parallel configurations from abinit autoparal.
    max_ncpus, min_eff = options.max_ncpus, options.min_eff
    print("Getting all autoparal confs up to max_ncpus: ",max_ncpus," with efficiency >= ",min_eff)

    pconfs = template.abiget_autoparal_pconfs(max_ncpus, autoparal=1, verbose=options.verbose)
    if options.verbose: print(pconfs)

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)

    omp_threads = 1
    for istwfk in [1, 2]:
        work = flowtk.Work()
        for conf in pconfs:
            mpi_procs = conf.mpi_ncpus
            if not options.accept_conf(conf, omp_threads): continue

            manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
            inp = template.new_with_vars(conf.vars, istwfk=istwfk)
            work.register_scf_task(inp, manager=manager)

        print("Found %d configurations" % len(work))
        flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return

    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""Duck-typing tests"""
from __future__ import annotations

import collections
import warnings
import numpy as np

from typing import Any


def is_string(s: Any) -> bool:
    """True if s behaves like a string (duck typing test)."""
    try:
        s + " "
        return True
    except Exception:
        return False


def is_intlike(obj: Any) -> bool:
    """
    True if obj represents an integer (floats such as 1.0 are included as well).
    """
    # isinstance(i, numbers.Integral)
    try:
        # This to get rid of warnings about casting complex to real.
        if np.iscomplexobj(obj) and np.isreal(obj):
            return int(obj.real) == obj
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                #print("hello", int(obj) == obj)
                return int(obj) == obj

    except (ValueError, TypeError) as exc:
        #print(exc)
        return False

    return False


def is_number_like(obj: Any) -> bool:
    """True if obj represents a number."""
    try:
        obj - 1
        return True
    except TypeError:
        return False


def is_listlike(obj: Any) -> bool:
    """True if obj is list-like."""
    #if isinstance(branch, (list, tuple, np.ndarray)):
    if isinstance(obj, np.ndarray): return True
    if not isinstance(obj, collections.abc.Sequence): return False
    if is_string(obj): return False

    try:
        obj[:0]
        return True
    except TypeError:
        return False


def list_ints(arg: Any) -> list:
    """
    Always return a list of int, given a int or list of integers as input.

    :Examples:

    >>> list_ints(1)
    [1]
    """
    l = np.array(arg, dtype=int)
    return [int(l)] if l.size == 1 else l.tolist()


def torange(obj: Any) -> range:
    """
    Convert obj into a range. Accepts integer, slice object  or any object
    with an __iter__ method. Note that an integer is converted into range(int, int+1)

    >>> list(torange(1))
    [1]
    >>> list(torange(slice(0, 4, 2)))
    [0, 2]
    >>> list(torange([1, 4, 2]))
    [1, 4, 2]
    """
    if is_intlike(obj):
        return range(obj, obj + 1)

    elif isinstance(obj, slice):
        start = obj.start if obj.start is not None else 0
        step = obj.step if obj.step is not None else 1
        return range(start, obj.stop, step)

    else:
        try:
            return obj.__iter__()
        except Exception:
            raise TypeError("Don't know how to convert %s into a range object" % str(obj))


def as_slice(obj: Any) -> slice:
    """
    Convert an integer, a string or a slice object into slice.

    >>> assert as_slice(5) == slice(5, 6, 1)
    >>> assert as_slice("[1:4]") == slice(1, 4, 1)
    >>> assert as_slice("1::2") == slice(1, None, 2)
    """
    if isinstance(obj, slice) or obj is None: return obj

    try:
        # integer.
        if int(obj) == float(obj): return slice(int(obj), int(obj)+1, 1)
    except Exception:
        # assume string defining a python slice [start:stop:step]
        if not obj: return None
        if obj.count("[") + obj.count("]") not in (0, 2):
            raise ValueError("Invalid string %s" % obj)

        obj = obj.replace("[", "").replace("]", "")
        n = obj.count(":")
        if n == 0:
            obj = int(obj)
            return slice(obj, obj+1)

        tokens = [int(f) if f else None for f in obj.split(":")]
        if len(tokens) == 2: tokens.append(1)
        if tokens[2] is None: tokens[2] = 1

        return slice(*tokens)

    raise ValueError("Cannot convert %s into a slice:\n%s" % (type(obj), obj))


class NoDefaultProvided:
    pass


def hasattrd(obj: Any, name: str) -> bool:
    """
    The arguments are an object and a string.
    The result is True if the string is the name of one of the object’s attributes, False if not.
    Unlike the builtin hasattr, hasattrd supports dot notation e.g. hasattr(int, "__class__.__name__")
    (This is implemented by calling getattrd(object, name) and seeing whether it raises an exception or not.)
    """
    try:
        getattrd(obj, name)
        return True
    except AttributeError:
        return False


def getattrd(obj: Any, name: str, default=NoDefaultProvided) -> Any:
    """
    Same as getattr(), but allows dot notation lookup e.g. getattrd(obj, "a.b")

    Raises: AttributeError if ``name`` is not found and ``default`` is not given.

    Discussed in: http://stackoverflow.com/questions/11975781
    """
    from functools import reduce
    try:
        return reduce(getattr, name.split("."), obj)
    except AttributeError:
        if default is not NoDefaultProvided:
            return default
        raise
